import io
import logging
from idlib.utils import log as _ilog
from augpathlib.utils import log as _alog
from pyontutils.utils import makeSimpleLogger, python_identifier  # FIXME update imports
from pyontutils.utils import TZLOCAL, utcnowtz, isoformat, isoformat_safe
from sparcur.config import config

log = makeSimpleLogger('sparcur')
logd = log.getChild('data')
loge = log.getChild('export')

# set augpathlib log format to pyontutils (also sets all child logs)
_alog.removeHandler(_alog.handlers[0])
_alog.addHandler(log.handlers[0])
# idlib logs TODO move to pyontutils probably?
_ilog.removeHandler(_alog.handlers[0])
_ilog.addHandler(log.handlers[0])


class GetTimeNow:
    def __init__(self):
        self._start_time = utcnowtz()
        self._start_local_tz = TZLOCAL()  # usually PST PDT

    @property
    def _start_time_local(self):
        return self._start_time.astimezone(self._start_local_tz)

    @property
    def START_TIMESTAMP(self):
        return isoformat(self._start_time)

    @property
    def START_TIMESTAMP_SAFE(self):
        return isoformat_safe(self._start_time)

    @property
    def START_TIMESTAMP_LOCAL(self):
        return isoformat(self._start_time_local)

    @property
    def START_TIMESTAMP_LOCAL_SAFE(self):
        return isoformat_safe(self._start_time_local)


class SimpleFileHandler:
    _FIRST = object()
    def __init__(self, log_file_path, *logs, mimic=_FIRST):
        self.log_file_handler = logging.FileHandler(log_file_path.as_posix())
        if mimic is self._FIRST and logs:
            self.mimic(logs[0])
        elif mimic:
            self.mimic(mimic)

        for log in logs:
            self(log)

    def __call__(self, *logs_to_handle):
        for log in logs_to_handle:
            log.addHandler(self.log_file_handler)

    def mimic(self, log):
        self.log_file_handler.setFormatter(log.handlers[0].formatter)


def bind_file_handler(log_file):
    # FIXME the this does not work with joblib at the moment
    from idlib.utils import log as idlog
    from protcur.core import log as prlog
    from orthauth.utils import log as oalog
    from ontquery.utils import log as oqlog
    from augpathlib.utils import log as alog
    from pyontutils.utils import log as pylog

    sfh = SimpleFileHandler(log_file, log)
    sfh(alog, idlog, oalog, oqlog, prlog, pylog)


class _log:
    """ logging prevents nice ipython recurions error printing
        so rename this class to log when you need fake logging """
    @staticmethod
    def debug(nothing): pass
    @staticmethod
    def info(nothing): pass
    @staticmethod
    def warning(nothing): print(nothing)
    @staticmethod
    def error(nothing): pass
    @staticmethod
    def critical(nothing): pass


want_prefixes = ('TEMP', 'FMA', 'UBERON', 'PATO', 'NCBITaxon', 'ilxtr', 'sparc',
                 'BIRNLEX', 'tech', 'unit', 'ILX', 'lex',)


def is_list_or_tuple(obj):
    return isinstance(obj, list) or isinstance(obj, tuple)


def symlink_latest(dump_path, path, relative=True):
    """ relative to allow moves of the containing folder
        without breaking links """

    if relative:
        dump_path = dump_path.relative_path_from(path)

    if path.exists():
        if not path.is_symlink():
            raise TypeError(f'Why is {path.name} not a symlink? '
                            f'{path!r}')

        path.unlink()

    path.symlink_to(dump_path)



from .xml import xml
from .disco import disco
from .triples import (TriplesExportDataset,
                      TriplesExportIdentifierMetadata,
                      TriplesExportSummary)
from .core import Export, ExportXml, latest_ir



{'config-search-paths': ['{:user-config-path}/sparcur/config.yaml',],
 'auth-variables':
 {'export-path': {'default': '{:user-data-path}/sparcur/export',
                  'environment-variables': 'SPARCUR_EXPORT_PATH SPARC_EXPORTS EXPORT_PATH'},
  'cache-path': {'default': '{:user-cache-path}/sparcur',
                 'environment-variables': 'SPARCUR_CACHE_PATH CACHE_PATH'},
  'log-path': {'default': '{:user-log-path}/sparcur',
               'environment-variables': 'SPARCUR_LOG_PATH LOG_PATH'},
  'blackfynn-organization': {'environment-variables': 'BLACKFYNN_ORGANIZATION'},
  'hypothesis-api-key': {'environment-variables': 'HYP_API_KEY HYP_API_TOKEN'},
  'hypothesis-group': {'environment-variables': 'HYP_GROUP'},
  'hypothesis-user': {'environment-variables': 'HYP_USER'},
  'datasets-noexport': None,
  'datasets-sparse': None,
  'datasets-no': None,
  'sparse-limit': {'default': 10000,
                   'environment-variables': 'SPARCUR_SPARSE_LIMIT SPARSE_LIMIT'},}}



from pathlib import Path
from flask import Flask, request, url_for
import htmlfn as hfn
from htmlfn import htmldoc, atag
from htmlfn import table_style, navbar_style
from pyontutils import clifun as clif
from sparcur import pipelines as pipes
from sparcur.curation import Integrator
from sparcur.utils import log

log = log.getChild('server')

clif.Dispatcher.url_for = staticmethod(url_for)


def nowrap(class_, tag=''):
    return (f'{tag}.{class_}'
            '{ white-space: nowrap; }')


def wrap_tables(*tables, title=None):
    return htmldoc(*tables,
                   styles=(table_style, nowrap('td', 'col-id')),
                   title=title)


def get_dataset_index(project_path):
    {c.cache.id:c for c in project_path.children}

    
def make_app(report, project_path, name='spc-server'):
    app = Flask(name)
    yield app

    bp = '/dashboard'

    @app.route(f'{bp}/datasets')
    def route_datasets(id):
        table, title = report._print_paths(project_path.children)
        return wrap_tables(table, title=title)

    @app.route(f'{bp}/datasets/<id>')
    @app.route(f'{bp}/datasets/<id>/ttl')
    @app.route(f'{bp}/datasets/<id>/json')
    def route_datasets_id(id, ext=None):
        dataset_index = get_dataset_index(project_path)
        if id not in dataset_index:
            return abort(404)

        dataset = dataset_index[id]
        tables = []
        try:
            ddt = [['TO', 'DO'], [id, 'derive tables from curation export!']]
            table, _ = report._print_table(ddt)
            tables.append(table)
        except StopIteration:
            return abort(404)  # FIXME ... no data instead plus iterate

        return wrap_tables(*tables, title='Dataset metadata tables')

    @app.route(f'{bp}/reports')
    @app.route(f'{bp}/reports/')
    def route_reports():
        report_names = (
            'completeness',
            'size',
            'filetypes',
            'pathids',
            'keywords',
            'samples',
            'subjects',
            'errors',
            'terms',
            'contributors',
        )
        report_links = [atag(url_for(f'route_reports_{rn}', ext=None), rn) + '<br>\n'
                        for rn in report_names]
        return htmldoc('Reports<br>\n',
                       *report_links,
                       title='Reports')

    @app.route(f'{bp}/reports/completeness')
    @app.route(f'{bp}/reports/completeness<ext>')
    def route_reports_completeness(ext=wrap_tables):
        return report.completeness(ext=ext)

    @app.route(f'{bp}/reports/size')
    @app.route(f'{bp}/reports/size<ext>')
    def route_reports_size(ext=wrap_tables):
        return report.size(dirs=project_path.children, ext=ext)

    @app.route(f'{bp}/reports/filetypes')
    @app.route(f'{bp}/reports/filetypes<ext>')
    def route_reports_filetypes(ext=None):
        if ext is not None:  # TODO
            return 'Not found', 404

        tables = []
        for table, title in report.filetypes():
            tables.append(table + '<br>\n')

        return wrap_tables(*tables, title='Filetypes')

    @app.route(f'{bp}/reports/pathids')
    @app.route(f'{bp}/reports/pathids<ext>')
    def route_reports_pathids(ext=wrap_tables):
        return report.pathids(ext=ext)

    @app.route(f'{bp}/reports/keywords')
    @app.route(f'{bp}/reports/keywords<ext>')
    def route_reports_keywords(ext=wrap_tables):
        return report.keywords(ext=ext)

    @app.route(f'{bp}/reports/samples')
    @app.route(f'{bp}/reports/samples<ext>')
    def route_reports_samples(ext=wrap_tables):
        return report.samples(ext=ext)

    @app.route(f'{bp}/reports/subjects')
    @app.route(f'{bp}/reports/subjects<ext>')
    def route_reports_subjects(ext=wrap_tables):
        return report.subjects(ext=ext)

    @app.route(f'{bp}/reports/errors')
    @app.route(f'{bp}/reports/errors<ext>')
    def route_reports_errors(ext=wrap_tables):
        return 'TODO'
        table, title = report.errors()
        return wrap_tables(table, title=title)

    @app.route(f'{bp}/reports/errors/<id>')
    @app.route(f'{bp}/reports/errors/<id>.<ext>')
    def route_reports_errors_id(id, ext=wrap_tables):
        tables, formatted_title, title = report.errors(id=id)
        log.info(id)
        if tables is None:
            return 'Not found', 404
        return wrap_tables(formatted_title, *tables, title=title)

    @app.route(f'{bp}/reports/terms')
    @app.route(f'{bp}/reports/terms<ext>')
    def route_reports_terms(ext=None):
        if ext is not None:  # TODO
            return 'Not found', 404

        tables = []
        for table, title in report.terms():
            tables.append(hfn.h2tag(title) + '<br>\n')
            tables.append(table + '<br>\n')

        return wrap_tables(*tables, title='Terms')

    @app.route(f'{bp}/reports/contributors')
    @app.route(f'{bp}/reports/contributors<ext>')
    def route_reports_contributors(ext=None):
        return report.contributors(ext=ext)

    @app.route(f'{bp}/apinat/demo')
    @app.route(f'{bp}/apinat/demo<ext>')
    def route_apinat_demo(ext=None):
        source = Path('~/ni/sparc/apinat/sources/').expanduser()  # FIXME config probably
        rm = pipes.ApiNATOMY(source / 'apinatomy-resourceMap.json')
        r = pipes.ApiNATOMY_rdf(rm.data)  # FIXME ... should be able to pass the pipeline
        if ext == '.ttl':
            return r.data.ttl, 200, {'Content-Type': 'text/turtle; charset=utf-8',}

        return hfn.htmldoc(r.data.ttl_html,
                           styles=(hfn.ttl_html_style,),
                           title='ApiNATOMY demo')

    @app.route(f'{bp}/reports/access')
    @app.route(f'{bp}/reports/access<ext>')
    def route_reports_access(ext=wrap_tables):
        return report.access(ext=ext)

    @app.route(f'{bp}/run/datasets/')
    @app.route(f'{bp}/run/datasets/<id>')
    def route_run_datasets(id=None):
        # TODO permissioning
        if id is None:
            pass

        # TODO send a message to/fork a process to run an export of a specific dataset



import sqlite3


class MetaStore:
    """ A local backup against accidental xattr removal """
    _attrs = ('bf.id',
              'bf.file_id',
              'bf.size',
              'bf.created_at',
              'bf.updated_at',
              'bf.checksum',
              'bf.error')
    attrs = 'xattrs',
    # FIXME horribly inefficient 1 connection per file due to the async code ... :/
    def __init__(self, db_path):
        self.db_path = db_path
        self.setup()

    def conn(self):
        return sqlite3.connect(self.db_path.as_posix())

    def setup(self):
        if not self.db_path.parent.exists():
            self.db_path.parent.mkdir(parents=True)

        sqls = (('CREATE TABLE IF NOT EXISTS path_xattrs'
                 '('
                 'id TEXT PRIMARY KEY,'  # for hypothesis ids this can be string(??)
                 'xattrs BLOB'  # see path meta for the packed representation
                 ');'),
                ('CREATE UNIQUE INDEX IF NOT EXISTS path_xattrs_u_id ON path_xattrs (id);'))
        conn = self.conn()
        with conn:
            for sql in sqls:
                conn.execute(sql)

    def bulk(self, id_blobs):  # FIXME no longer a dict really ...
        cols = ', '.join(_.replace('.', '_') for _ in self.attrs)
        values_template = ', '.join('?' for _ in self.attrs)
        sql = ('INSERT OR REPLACE INTO path_xattrs '
               f'(id, {cols}) VALUES (?, {values_template})')
        conn = self.conn()
        with conn:
            for id, blob in id_blobs:
                conn.execute(sql, args)
            return
            for path, attrs in pdict.items():
                args = path.as_posix(), *self.convert_attrs(attrs)
                conn.execute(sql, args)

    def remove(self, path):
        sql = 'DELETE FROM path_xattrs WHERE id = ?'
        args = path.as_posix(),
        conn = self.conn()
        with conn:
            return conn.execute(sql, args)

    def convert_attrs(self, attrs):
        for key in self.attrs:
            if key in attrs:
                yield attrs[key]
            else:
                yield None

    def xattrs(self, path):
        sql = 'SELECT xattrs FROM path_xattrs WHERE id = ?'
        args = path.as_posix(),
        conn = self.conn()
        with conn:
            cursor = conn.execute(sql, args)
            blob = cursor.fetchone()
            if blob:
                return PathMeta.from_metastore(blob)
            #print(values)
            #if values:
                #return 
                #keys = [n.replace('_', '.', 1) for n, *_ in cursor.description]
                #print(keys, values)
                #return {k:v for k, v in zip(keys, values) if k != 'path' and v is not None}  # skip path itself
            #else:
                #return {}

    def setxattr(self, path, key, value):
        return self.setxattrs(path, {key:value})

    def setxattrs(self, path, attrs):
        # FIXME skip nulls on replace
        cols = ', '.join(attrs)
        values_template = ', '.join('?' for _ in self.attrs)
        sql = (f'INSERT OR REPLACE INTO path_xattrs (id, {cols}) VALUES (?, {values_template})')
        args = path.as_posix(), *self.convert_attrs(attrs)
        conn = self.conn()
        with conn:
            return conn.execute(sql, args)

    def getxattr(self, path, key):
        if key in self.attrs:
            col = key.replace('.', '_')
            sql = f'SELECT {col} FROM path_xattrs WHERE id = ?'
            args = path.as_posix(),
            conn = self.conn()
            with conn:
                return conn.execute(sql, args)
        else:
            print('WARNING unknown key', key)



import unittest
from sparcur.core import adops, DictTransformer
from sparcur.derives import Derives as De


class Examples:
    # yield expect, data, args
    @property
    def add(self):
        yield {'hello': 'world'}, {}, (['hello'], 'world')
        yield {'hello': {'there': {'general': 'world'}}}, {}, (['hello', 'there', 'general'], 'world')
        yield {'hello': {'there': {'general': 'world'}}}, {'hello': {}}, (['hello', 'there', 'general'], 'world')
        yield {'world': [{'hello': 'there'},
                         {'hello': 'there'}]}, {'world': [{},
                                                          {}]}, (['world', int, 'hello'], 'there')

    @property
    def add_error(self):
        yield {'hello':'1'}, {'hello':'1'}, (['hello'], 'there')
        yield {'hello':'2'}, {'hello':'2'}, ([], 'there')
        yield {'hello':'3'}, {'hello':'3'}, ('lolol', 'there')  # this might succeed :x indeed it did
        yield {'hello': {'there': {'general': 'world'}}}, {'hello': 'OH NO'}, (['hello', 'there', 'general'], 'world')

    @property
    def get(self):
        yield {'hello': 'world'}, {'hello','world'}, (['hello'],)
        yield {'hello': {'there': {'general': 'world'}}}, {}, (['hello', 'there', 'general'],)

    @property
    def get_error(self):
        yield {'hello': 'world'}, {'hello','world'}, (['another'],)

    @property
    def delete(self):
        yield {}, {'hello','world'}, (['hello'],)
        yield ({'hello': {'there': {'keep': 'MEEE'}}},
               {'hello': {'there': {'general': 'world', 'keep': 'MEEE'}}},
               (['hello', 'there', 'general'],))

    @property
    def delete_error(self):
        yield {'hello': 'world'}, {'hello','world'}, (['hello', 'world'],)

    @property
    def copy(self):
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['hello'], ['another'])

    @property
    def copy_error(self):
        yield {'another':'world', 'hello': 'world'}, {'another':'world', 'hello': 'world'}, (['hello'], ['another'])
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, ([], ['another'])
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['hello'], [])  # dicts are closed world can't break on through
        yield {'another':'world', 'hello': 'world'}, {'hello': 'world'}, (['no-source'], ['oops'])

    @property
    def move(self):
        yield {'another':'world'}, {'hello': 'world'}, (['hello'], ['another'])

    @property
    def move_error(self):
        yield {'another':'1'}, {}, (['hello'], ['another'])
        yield {'another':'2'}, {'another':0, 'hello':'world'}, (['hello'], ['another'])


class TestDer(unittest.TestCase):
    def test_der(self):
        assert De.contributor_name('b, a') == ('a', 'b')


class ExamplesDT:

    @property
    def derive(self):
        yield ({'a': 2, 'b': 3, 'c': 5, 'd': 6, 'e': 1},
               {'a': 2, 'b': 3}, [[[['a'], ['b']],
                                   lambda a, b: (a + b, a * b),
                                   [['c'], ['d']]],
                                  # look you can chain them!
                                  [[['c'], ['d']],
                                   lambda c, d: (d - c,),
                                   [['e']]]])
        yield ({'name': 'b, a', 'first_name': 'a', 'last_name': 'b'},
               {'name': 'b, a'}, [[[['name']],
                                   De.contributor_name,
                                   [['first_name'], ['last_name']]]])

        # this is no longer an error because we allow TypeError and
        # catch the issue using the schema
        yield ({'a': 1}, {'a': 1}, [[[['a']],
                                     # TODO is there any way to catch this information early ?
                                     lambda _: None,  # previously None would include a TypeError
                                     [['b']]]])

    @property
    def lift(self):
        yield {'hello':'there'}, {'hello': 'world'}, [[['hello'], lambda v: 'there']]


def failed_to_error(function, data, args):
    try:
        function(data, *args)
        print(function, data, args)
        return True
    except BaseException as e:
        return False


class Populator:
    apply = True
    @classmethod
    def populate(cls):
        for name in cls.functions:
            name_error = name + '_error'
            name_negative = name + '_negative'

            test_name = 'test_' + name
            function_to_test = getattr(cls.to_test, name)
            def fun(self, name=name, function=function_to_test):
                bads = [(expect, data, args)
                        for expect, data, args in getattr(self, name)
                        if not function(data, *(args if cls.apply else (args,))) and data != expect]
                assert not bads, bads

            setattr(cls, test_name, fun)

            if hasattr(cls, name_error):
                test_name_error = 'test_' + name_error
                def efun(self, name=name_error, function=function_to_test):
                    bads = [(expect, data, args)
                            for expect, data, args in getattr(self, name)
                            if failed_to_error(function, data, (args if cls.apply else (args,)))]
                    assert not bads, bads

                setattr(cls, test_name_error, efun)

            if hasattr(cls, name_negative):
                test_name_negative = 'test_' + name_negative
                def nfun(self, name=name_negative, function=function_to_test):
                    bads = [(expect, data, args)
                            for expect, data, args in getattr(self, name)
                            if not function(data, *(args if cls.apply else (args,))) and data == expect]
                    assert not bads, bads

                setattr(cls, test_name_negative, nfun)


class TestAdops(Examples, Populator, unittest.TestCase):
    functions = 'add', 'copy', 'move'
    to_test = adops
TestAdops.populate()


class TestDictTransformer(ExamplesDT, Populator, unittest.TestCase):
    functions = 'lift', 'derive'
    to_test = DictTransformer
    apply = False
TestDictTransformer.populate()



from urllib.parse import quote
import idlib
import rdflib
from pyontutils.core import OntGraph
from pyontutils.utils import isoformat, utcnowtz
from pyontutils.namespaces import (TEMP,
                                   isAbout,
                                   sparc,
                                   rdf,
                                   rdfs,
                                   owl,
                                   skos,
                                   dc,)
from sparcur import converters as conv
from sparcur.core import (adops,
                          OntCuries,
                          curies_runtime,)
from sparcur.utils import loge, log
from sparcur.protocols import ProtcurData


class TriplesExport(ProtcurData):

    def __init__(self, data_json, *args, teds=tuple(), **kwargs):
        self.data = data_json
        self.teds = teds
        self.id = self.data['id']
        self.uri_api = self.data['meta']['uri_api']
        self.uri_human = self.data['meta']['uri_human']
        self.folder_name = self.data['meta']['folder_name']
        #self.title = self.data['meta']['title'] if 

    @property
    def timestamp_export_start(self):
        return adops.get(self.data, ['prov', 'timestamp_export_start'])

    @property
    def protocol_uris(self):  # FIXME this needs to be pipelined
        try:
            yield from adops.get(self.data, ['meta', 'protocol_url_or_doi'])
        except exc.NoSourcePathError:
            pass

    @property
    def ontid(self):
        return rdflib.URIRef(f'https://cassava.ucsd.edu/sparc/ontologies/{self.id}')

    @property
    def header_graph_description(self):
        raise NotImplementedError

    @property
    def triples_header(self):
        ontid = self.ontid
        nowish = utcnowtz()
        epoch = nowish.timestamp()
        iso = isoformat(nowish)
        ver_ontid = rdflib.URIRef(ontid + f'/version/{epoch}/{self.id}')
        sparc_methods = rdflib.URIRef('https://raw.githubusercontent.com/SciCrunch/'
                                      'NIF-Ontology/sparc/ttl/sparc-methods.ttl')

        pos = (
            (rdf.type, owl.Ontology),
            (owl.versionIRI, ver_ontid),
            (owl.versionInfo, rdflib.Literal(iso)),
            (isAbout, rdflib.URIRef(self.uri_api)),
            (TEMP.hasHumanUri, rdflib.URIRef(self.uri_human)),
            (rdfs.label, rdflib.Literal(f'{self.folder_name} curation export graph')),
            (rdfs.comment, self.header_graph_description),
            (owl.imports, sparc_methods),
            (TEMP.TimestampExportStart, rdflib.Literal(self.timestamp_export_start)),
        )
        for p, o in pos:
            yield ontid, p, o

    @property
    def graph(self):
        """ you can populate other graphs, but this one runs once """
        if not hasattr(self, '_graph'):
            graph = OntGraph()
            self.populate(graph)
            self._graph = graph

        return self._graph

    @property
    def ttl(self):
        return self.graph.serialize(format='nifttl')

    def populate(self, graph):
        def warn(triple):
            for element in triple:
                if (not (isinstance(element, rdflib.URIRef) or
                         isinstance(element, rdflib.BNode) or
                         isinstance(element, rdflib.Literal)) or
                    (hasattr(element, '_value') and (isinstance(element._value, dict) or
                                                     isinstance(element._value, list) or
                                                     isinstance(element._value, tuple))) or
                    (isinstance(element, rdflib.URIRef) and (element.startswith('<') or
                                                             not rdflib.term._is_valid_uri(element)))):
                    #if (isinstance(element, rdflib.URIRef) and element.startswith('<')):
                        #breakpoint()
                    loge.critical(element)

            return triple

        OntCuries.populate(graph)  # ah smalltalk thinking
        if hasattr(self, 'uri_api'):
            base = self.uri_api + '/'
            graph.namespace_manager.populate_from(curies_runtime(base))

        [graph.add(t) for t in self.triples_header if warn(t)]
        for t in self.triples:
            if warn(t):
                graph.add(t)


class TriplesExportSummary(TriplesExport):
    def __iter__(self):
        yield from self.data['datasets']

    @property
    def header_graph_description(self):
        return rdflib.Literal(f'SPARC organization graph for {self.id}')

    @property
    def triples(self):
        if self.teds:
            # TODO conjuctive graph?
            for ted in self.teds:
                # FIXME BNode collision risk? Probably not?
                # FIXME yeah, figuring out conjuctive graph for header plus
                sont = next(ted.graph[:rdf.type:owl.Ontology])
                for s, p, o in ted.graph:
                    if s == sont:
                        continue

                    yield s, p, o

        else:
            for dataset_blob in self:
                ted = TriplesExportDataset(dataset_blob)
                yield from ted.triples


class TriplesExportDataset(TriplesExport):
    @property
    def header_graph_description(self):
        return rdflib.Literal(f'SPARC single dataset graph for {self.id}')

    @property
    def subjects(self):
        if 'subjects' in self.data:
            yield from self.data['subjects']

    @property
    def samples(self):
        if 'samples' in self.data:
            yield from self.data['samples']

    @property
    def dsid(self):
        return rdflib.URIRef(self.uri_api)

    def ddt(self, data):
        if 'contributors' in data:
            if 'creators' in data:
                creators = data['creators']
            else:
                creators = []
            for c in data['contributors']:
                creator = ('name' in c and
                           [cr for cr in creators if 'name' in cr and cr['name'] == c['name']])
                yield from self.triples_contributors(c, creator=creator)

    def triples_contributors(self, contributor, creator=False):
        try:
            dsid = self.dsid  # FIXME json reload needs to deal with this
        except BaseException as e:  # FIXME ...
            loge.exception(e)
            return

        cid = contributor['id']

        if isinstance(cid, idlib.Stream):  # FIXME nasty branch
            s = cid.asUri(rdflib.URIRef)
        elif isinstance(cid, dict):
            if isinstance(cid['id'], idlib.Stream):  # FIXME nasty branch
                s = cid['id'].asUri(rdflib.URIRef)
            else:
                raise NotImplementedError(f'{type(cid["id"])}: {cid["id"]}')
        else:
            s = rdflib.URIRef(cid)  # FIXME json reload needs to deal with this

        if 'blackfynn_user_id' in contributor:
            userid = rdflib.URIRef(contributor['blackfynn_user_id'])
            yield s, TEMP.hasBlackfynnUserId, userid

        yield s, rdf.type, owl.NamedIndividual
        yield s, rdf.type, sparc.Person
        yield s, TEMP.contributorTo, dsid  # TODO other way around too? hasContributor
        converter = conv.ContributorConverter(contributor)
        yield from converter.triples_gen(s)
        if creator:
            yield s, TEMP.creatorOf, dsid

        # dataset <-> contributor object
        dcs = rdflib.BNode()

        yield dcs, rdf.type, owl.NamedIndividual
        yield dcs, rdf.type, sparc.DatasetContribution
        yield dcs, TEMP.aboutDataset, dsid  # FIXME forDataset?
        yield dcs, TEMP.aboutContributor, s
        dconverter = conv.DatasetContributorConverter(contributor)
        for _s, p, o in dconverter.triples_gen(dcs):
            if p == sparc.isContactPerson and o._value == True:
                yield dsid, TEMP.hasContactPerson, s
            yield _s, p, o

    @property
    def triples(self):
        # FIXME ick
        data = self.data
        try:
            dsid = self.uri_api
        except BaseException as e:  # FIXME ...
            raise e
            return

        if 'meta' in data:
            meta_converter = conv.MetaConverter(data['meta'], self)
            yield from meta_converter.triples_gen(dsid)
        else:
            loge.warning(f'{self} has no meta!')  # FIXME split logs into their problems, and our problems

        if 'submission' in data:
            submission_converter = conv.SubmissionConverter(data['submission'], self)
            yield from submission_converter.triples_gen(dsid)
        else:
            loge.warning(f'{self} has no submission!')  # FIXME split logs into their problems, and our problems

        if 'status' not in data:
            breakpoint()

        yield from conv.StatusConverter(data['status'], self).triples_gen(dsid)

        #converter = conv.DatasetConverter(data)
        #yield from converter.triples_gen(dsid)

        def id_(v):
            s = rdflib.URIRef(dsid)
            yield s, rdf.type, owl.NamedIndividual
            yield s, rdf.type, sparc.Dataset
            yield s, rdfs.label, rdflib.Literal(self.folder_name)  # not all datasets have titles

        yield from id_(self.id)

        #for subjects in self.subjects:
            #for s, p, o in subjects.triples_gen(subject_id):
                #if type(s) == str:
                    #breakpoint()
                #yield s, p, o

        yield from self.ddt(data)
        yield from self.triples_subjects
        yield from self.triples_samples

    def subject_id(self, v, species=None):  # TODO species for human/animal
        if isinstance(v, int):
            loge.critical('darn it max normlize your ids!')
            v = str(v)

        v = quote(v, safe=tuple())
        s = rdflib.URIRef(self.dsid + '/subjects/' + v)
        return s

    @property
    def triples_subjects(self):
        try:
            dsid = self.dsid  # FIXME json reload needs to deal with this
        except BaseException as e:  # FIXME ...
            loge.exception(e)
            return

        def triples_gen(prefix_func, subjects):

            for i, subject in enumerate(subjects):
                converter = conv.SubjectConverter(subject)
                if 'subject_id' in subject:
                    s_local = subject['subject_id']
                else:
                    s_local = f'local-{i + 1}'  # sigh

                s = prefix_func(s_local)
                yield s, rdf.type, owl.NamedIndividual
                yield s, rdf.type, sparc.Subject
                yield s, TEMP.hasDerivedInformationAsParticipant, dsid
                yield dsid, TEMP.isAboutParticipant, s
                yield from converter.triples_gen(s)
                continue
                for field, value in subject.items():
                    convert = getattr(converter, field, None)
                    if convert is not None:
                        yield (s, *convert(value))
                    elif field not in converter.known_skipped:
                        loge.warning(f'Unhandled subject field: {field}')

        yield from triples_gen(self.subject_id, self.subjects)

    def primary_key(self, v, species=None):  # TODO species for human/animal
        #v = v.replace(' ', '%20')  # FIXME use quote urlencode
        v = quote(v, safe=tuple())
        s = rdflib.URIRef(self.dsid + '/samples/' + v)
        return s

    @property
    def triples_samples(self):
        try:
            dsid = self.dsid  # FIXME json reload needs to deal with this
        except BaseException as e:  # FIXME ...
            loge.exception(e)
            return

        conv.SampleConverter._subject_id = self.subject_id  # FIXME
        conv.SampleConverter.dsid = self.dsid  # FIXME FIXME very evil
        # yes this indicates that converters and exporters are
        # highly related here ...
        def triples_gen(prefix_func, samples):
            for i, sample in enumerate(samples):
                converter = conv.SampleConverter(sample)
                if 'primary_key' in sample:
                    s_local = sample['primary_key']
                else:
                    s_local = f'local-{i + 1}'  # sigh

                s = prefix_func(s_local)
                yield s, rdf.type, owl.NamedIndividual
                yield s, rdf.type, sparc.Sample
                yield s, TEMP.hasDerivedInformationAsParticipant, dsid  # domain particiant range information artifact
                # specimen - participant: -> process instance - ilxtr:hasInformationOutput -> data files - partOf: -> dataset
                # collapses to? specimen - hasInformationDerivedFromProce -> <- containsInformationAbout - dataset
                yield dsid, TEMP.isAboutParticipant, s  # containsInformationAboutParticipant[Primary] TEMP.containsInformationAbout, isAbout is probably a better base
                # could be further refined to isAboutParticiantPrimary, with a note that if multiple measurement processes happened, there can be multiple primaries for a dataset
                yield from converter.triples_gen(s)
                # see https://github.com/information-artifact-ontology/IAO/issues/60, there isn't a good inverse relation
                # original though was subjectOfInformation, but that was confusing in the current terminology where subject already has 2 meanings
                # hasInformationDerivedFromProcessWhereWasParticipant -> hasInformationDerivedFromProcessWhereWasPrimaryParticipant seems most correct, but is extremely verbose
                # hasDerivedInformationAsParticipant -> hasDerivedInformationAsParticipantPrimary materialize the role into the predicate? seems reasonable
                continue
                for field, value in sample.items():
                    convert = getattr(converter, field, None)
                    if convert is not None:
                        yield (s, *convert(value))
                    elif field not in converter.known_skipped:
                        loge.warning(f'Unhandled sample field: {field}')

        yield from triples_gen(self.primary_key, self.samples)


class TriplesExportIdentifierMetadata(TriplesExport):
    def __init__(self, data_json, *args, **kwargs):
        self.data = data_json
        self.id = self.data['id']

    @property
    def header_graph_description(self):
        return rdflib.Literal(f'Additional metadata for identifiers referenced in export.')

    @property
    def triples_header(self):
        ontid = self.ontid
        nowish = utcnowtz()
        epoch = nowish.timestamp()
        iso = isoformat(nowish)
        ver_ontid = rdflib.URIRef(ontid + f'/version/{epoch}/{self.id}')
        pos = (
            (rdf.type, owl.Ontology),
            (owl.versionIRI, ver_ontid),
            (owl.versionInfo, rdflib.Literal(iso)),
            (rdfs.comment, self.header_graph_description),
            (TEMP.TimestampExportStart, rdflib.Literal(self.timestamp_export_start)),
        )
        for p, o in pos:
            yield ontid, p, o

    def published_online(self, blob):
        try:
            dpl = blob['issued']['date-parts']
        except KeyError as e:
            log.critical(e)
            return None

        dp = dpl[0]
        if len(dp) == 3:
            y, m, d = dp
            return f'{y}-{m:0>2}-{d:0>2}'
        elif len(dp) == 2:
            y, m = dp
            return f'{y}-{m:0>2}'
        elif len(dp) == 1:
            y = dp
            return f'{y}'
        else:
            raise NotImplementedError(f'what the? {dp}')

    @property
    def triples(self):
        crossref_doi_pred = rdflib.term.URIRef('http://prismstandard.org/namespaces/basic/2.1/doi')
        for blob in self.data['identifier_metadata']:
            id = blob['id']
            if not isinstance(id, idlib.Stream):
                id = idlib.Auto(id)

            if not hasattr(id, 'asUri'):
                breakpoint()

            s = id.asUri(rdflib.URIRef)
            if 'source' in blob:
                source = blob['source']  # FIXME we need to wrap this in our normalized representation
                if source == 'Crossref':  # FIXME CrossrefConvertor etc. OR put it in idlib as a an alternate ttl
                    pos = (
                        (rdf.type, owl.NamedIndividual),
                        (rdf.type, TEMP[blob['type']]),
                        (dc.publisher, blob['publisher']),
                        #(dc.type, blob['type']),  # FIXME semantify
                        (dc.title, blob['title']),
                        (dc.date, self.published_online(blob)),  # FIXME .... dangerzone
                    )
                    g = OntGraph()
                    doi = idlib.Doi(id) if not isinstance(id, idlib.Doi) else id  # FIXME idlib streams need to recognize their own type in __new__
                    data = doi.ttl()
                    if data is None:  # blackfynn has some bad settings on their doi records ...
                        return

                    try:
                        g.parse(data=data, format='ttl')  # FIXME network bad
                    except BaseException as e:
                        loge.exception(e)

                    _tr = [s for s, p, o in g if p == crossref_doi_pred]
                    if _tr:
                        _their_record_s = _tr[0]
                        yield s, owl.sameAs, _their_record_s
                        yield from g
                    else:
                        g.debug()
                        log.critical('No crossref doi section in graph!')
                else:
                    msg = f'dont know what to do with {source}'
                    log.error(msg)
                    #raise NotImplementedError(msg)
                    return
            else:
                msg = f'dont know what to do with {blob} for {id.identifier}'
                log.error(msg)
                #raise NotImplementedError(msg)
                return

            for p, oraw in pos:
                if oraw is not None:
                    o = rdflib.Literal(oraw) if not isinstance(oraw, rdflib.URIRef) else oraw
                    yield s, p, o



import pathlib
import idlib
import rdflib
import dicttoxml
from pysercomb.pyr.types import ProtcurExpression, Quantity
from sparcur.core import OntTerm, get_all_errors
from sparcur.utils import loge, is_list_or_tuple


def xml(dataset_blobs):
    #datasets = []
    #contributors = []
    subjects = []
    errors = []
    resources = []

    def normv(v):
        if is_list_or_tuple(v):
            return [normv(_) for _ in v]
        if isinstance(v, dict):
            return {k:normv(v) for k, v in v.items()}
        if isinstance(v, str) and v.startswith('http'):
            # needed for loading from json that has been serialized
            # rather than from our internal representation
            # probably better to centralized the reload ...
            v = OntTerm(v)
            return v.asCell()
        if isinstance(v, rdflib.URIRef):  # FIXME why is this getting converted early?
            ot = OntTerm(v)
            return ot.asCell()
        if isinstance(v, ProtcurExpression):
            return str(v)  # FIXME for xml?
        if isinstance(v, Quantity):
            return str(v)
        elif isinstance(v, pathlib.Path):
            return str(v)
        elif isinstance(v, idlib.Stream):
            return v.asCell()
        #elif isinstance(v, list) or isinstance(v, str):
            #return v
        else:
            #loge.debug(repr(v))
            return v

    for dataset_blob in dataset_blobs:
        id = dataset_blob['id']
        dowe = dataset_blob
        #id = dataset.id
        #dowe = dataset.data
        if 'subjects' in dowe:
            for subject in dowe['subjects']:
                subject['dataset_id'] = id
                subject = {k:normv(v) for k, v in subject.items()}
                subjects.append(subject)

        if 'resources' in dowe:
            for res in dowe['resources']:
                res['dataset_id'] = id
                res = {k:normv(v) for k, v in res.items()}
                resources.append(res)

        if 'errors' in dowe:
            ers = get_all_errors(dowe)
            for er in ers:
                if er['pipeline_stage'] == 'SPARCBIDSPipeline.data':
                    continue

                er['dataset_id'] = id
                er = {k:normv(v) for k, v in er.items()}
                errors.append(er)

    xs = dicttoxml.dicttoxml({'subjects': subjects})
    xr = dicttoxml.dicttoxml({'resources': resources})
    xe = dicttoxml.dicttoxml({'errors': errors})
    return (('subjects', xs),
            ('resources', xr),
            ('errors', xe))



import os
import json
import unittest
import pytest
from pyontutils.utils import Async, deferred
from sparcur.core import JEncode
from sparcur.paths import BlackfynnCache, Path
from sparcur.backends import BlackfynnRemote
from sparcur.config import auth
from sparcur.extract import xml as exml
from .common import path_project_container, examples_root

export = False

@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class TestExtractMetadata(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.organization_id = auth.get('blackfynn-organization')
        cls.BlackfynnRemote = BlackfynnRemote._new(Path, BlackfynnCache)
        cls.BlackfynnRemote.init(cls.organization_id)
        cls.anchor = cls.BlackfynnRemote.smartAnchor(path_project_container)
        cls.anchor.local_data_dir_init()
        cls.datasets = list(cls.anchor.children)

    def test_new_mbf_format(self):
        x = examples_root / 'mbf-example.xml'
        embf = exml.ExtractXml(x)
        d = embf.asDict()
        errors = d.pop('errors') if 'errors' in d else tuple()
        error_types = set(e['validator'] for es in errors for e in es)
        assert error_types == {'not'} or not error_types, f'unexpected error type! {error_types}'

    def test_mbf_header(self):
        test_id = 'N:dataset:bec4d335-9377-4863-9017-ecd01170f354'
        test_dataset = [d for d in self.datasets if d.id == test_id][0]
        if not list(test_dataset.local.children):
            rchilds = list(test_dataset.rchildren)
            xmls = [c for c in rchilds if c.suffix == '.xml']
            Async(rate=5)(deferred(x.fetch)() for x in xmls if not x.exists())
            #[x.fetch() for x in xmls if not x.exists()]
            local_xmls = [x.local for x in xmls]
        else:
            local_xmls = list(test_dataset.local.rglob('*.xml'))
            if any(p for p in local_xmls if not p.exists()):
                raise BaseException('unfetched children')

        embfs = [exml.ExtractXml(x) for x in local_xmls]
        d = embfs[0].asDict()
        blob = [e.asDict() for e in embfs]
        errors = [b.pop('errors') for b in blob if 'errors' in b]
        error_types = set(e['validator'] for es in errors for e in es)
        if export:
            with open('mbf-test.json', 'wt') as f:
                json.dump(blob, f, indent=2, cls=JEncode)
            with open('mbf-errors.json', 'wt') as f:
                json.dump(errors, f, indent=2, cls=JEncode)

        assert error_types == {'not'} or not error_types, f'unexpected error type! {error_types}'



import tempfile
from pathlib import Path
import orthauth as oa

auth = oa.configure_here('auth-config.py', __name__)


class config:
    organ_html_path = Path('../resources/sparc-nervous-system-graphic.html')  # FIXME include in distribution ...



import pprint
from typing import Tuple
from functools import wraps
from sparcur import schemas as sc
from sparcur import normalization as nml
from sparcur.core import log, logd, JPointer


def collect(*oops, unpacked=True):
    def decorator(generator_function):
        @wraps(generator_function)
        def inner(*args, **kwargs):
            out = tuple(generator_function(*args, **kwargs))
            if unpacked:
                return out
            else:
                return out,

        return inner

    if oops:
        generator_function, = oops
        return decorator(generator_function)
    else:
        return decorator


class Derives:
    """ all derives must return as a tuple, even if they are length 1 """

    @staticmethod
    def contributor_name(name) -> Tuple[str, str]: 
        if ',' in name:
            last, first = name.split(', ', 1)
        elif ' ' in name:
            first, last = name.split(' ', 1)
        else:
            first = 'ERROR IN NAME see lastName for content'
            last = name

        return first, last

    @staticmethod
    @collect
    def creators(contributors):
        for c in contributors:
            if 'contributor_role' in c and 'Creator' in c['contributor_role']:
                # FIXME diry diry mutation here that should happen in a documente way
                cont = {**c}  # instead of importing copy since this is a one deep
                cont.pop('contributor_role', None)
                cont.pop('is_contact_person', None)
                yield cont

    @staticmethod
    def award_number(raw_award_number, funding) -> str:
        return nml.NormAward(nml.NormAward(raw_award_number))

    @staticmethod
    def pi(contributors):
        pointers = [JPointer(f'/contributors/{i}')
                    for i, c in enumerate(contributors)
                    if (('PrincipalInvestigator' in c['contributor_role'])
                        if 'contributor_role' in c else False)]

        if len(pointers) == 1:
            pointer, = pointers
            return pointer

        return pointers

    @staticmethod
    def dataset_species(subjects) -> Tuple[tuple]:
        out = set()
        for subject in subjects:
            if 'species' in subject:
                out.add(subject['species'])

        if len(out) == 1:
            return next(iter(out))

        return tuple(out)



import idlib
import rdflib
import requests
from idlib.cache import cache, COOLDOWN
from orthauth.utils import QuietDict
from pyontutils import combinators as cmb
from pyontutils.namespaces import OntCuries, makeNamespaces, TEMP, isAbout, ilxtr
from pyontutils.namespaces import rdf, rdfs, owl, skos, dc
from sparcur import datasets as dat
from sparcur.core import log, logd, OntTerm, OntId, sparc
from sparcur.paths import Path
from sparcur.utils import log, logd
from sparcur.config import config, auth


class ProtcurData:

    def __init__(self, *args, **kwargs):
        super().__init__()  # this is the last in the chain atm

    def __call__(self, protocol_uri):
        """ can't use __call__ in subclasses where it is overwritten ... """
        yield from self.protcur_nodes(protocol_uri)

    def _protcur(self, protocol_uri, filter=lambda p: True):
        self.lazy_setup()
        protocol_uri = idlib.get_right_id(protocol_uri)
        gen = (p for p in self.protc if p.uri.startswith(protocol_uri.identifier) and filter(p))

        try:
            p = next(gen)
            yield p
            yield from gen
        except StopIteration:
            log.error(f'could not find annotations for {protocol_uri.identifier}')
            return

        if p.document.otherVersionUri:  # FIXME also maybe check /abstract?
            other_uri = p.document.otherVersionUri
            yield from (p for p in self.protc if p.uri.startswith(other_uri) and filter(p))

    @property
    def protcur(self):
        """ protcur nodes without parents """
        for protocol_uri in self.protocol_uris:
            yield from self._protcur(protocol_uri, filter=lambda p: not p.hasAstParent)

    @property
    def protcur_all(self):
        for protocol_uri in self.protocol_uris:
            yield from self._protcur(protocol_uri)

    def triples_protcur(self, protocol_subject):
        ps = list(self._protcur(str(protocol_subject)))
        anatomy = [(p, OntId('UBERON:' + str(p).split('UBERON:', 1)[-1].split(' ', 1)[0])) for p in ps
                   if p.astType == 'protc:input' and '(protc:input (term UBERON' in str(p)]
        #breakpoint()
        dataset_subject = rdflib.URIRef(self.uri_api)
        yield protocol_subject, TEMP.hasNumberOfProtcurAnnotations, rdflib.Literal(len(ps))
        done = set()
        for anno, term in anatomy:
            if term in done:
                continue

            done.add(term)
            yield from OntTerm(term).triples_simple
            o = term.u
            t = dataset_subject, TEMP.involvesAnatomicalRegion, o
            sl = rdflib.URIRef(anno.shareLink)
            av = (((ilxtr.annotationValue, rdflib.Literal(anno.value)),)
                  if anno.value != o else tuple())
            notes = [(ilxtr.curatorNote, rdflib.Literal(n)) for n in anno.curatorNotes]
            prov = [(ilxtr.hasAnnotation, sl)]
            yield t
            yield from cmb.annotation(t, *av, *notes, *prov)()

    @classmethod
    def lazy_setup(cls):
        """ Do not use setup since it is always run by Integrator
            We want lazy loading, and should probably conver all
            setup style things that hit the network to be lazy. """
        if not hasattr(cls, '_setup_ok'):
            cls.populate_annos()
            cls._setup_ok = True

    @staticmethod
    def populate_annos():
        from protcur.core import annoSync
        from protcur.analysis import Hybrid, protc
        from hyputils.hypothesis import HypothesisHelper, group_to_memfile
        from hyputils import hypothesis as hyp
        ProtcurData.protc = protc
        group = auth.get('hypothesis-group')
        get_annos, annos, stream_thread, exit_loop = annoSync(group_to_memfile(group + 'sparcur'),
                                                              helpers=(HypothesisHelper, Hybrid, protc),
                                                              group=group,
                                                              sync=False)

        # FIXME hack to workaround bad api key init for hyutils until we can integrate orthauth
        get_annos.api_token = auth.get('hypothesis-api-key')
        annos.clear()
        annos.extend(get_annos())

        # reset classes in case some other class has populated them
        # (e.g. during testing) FIXME this is a bad hack
        protc.reset()
        Hybrid.reset()

        # FIXME this is expensive and slow to continually recompute
        [protc(a, annos) for a in annos]
        [Hybrid(a, annos) for a in annos]

    def __lol__(self, protocol_uri):
        self.protocol_uri = protocol_uri


class ProtocolData(dat.HasErrors):
    # this class is best used as a helper class not as a __call__ class

    _instance_wanted_by = idlib.Pio, idlib.PioUser

    def __init__(self, id=None):  # FIXME lots of ways to use this class ...
        self.id = id  # still needed for the converters use case :/
        # FIXME protocol data shouldn't need do know anything about
        # what dataset is using it, >_<
        super().__init__(pipeline_stage=self.__class__)

    def protocol(self, uri):
        return self._get_protocol_json(uri)

    __call__ = protocol

    @property
    def protocol_uris(self):
        raise NotImplementedError('implement in subclass')
        yield

    @property
    def protocol_uris_resolved(self):
        if not hasattr(self, '_c_protocol_uris_resolved'):
            self._c_protocol_uris_resolved = list(self._protocol_uris_resolved)

        return self._c_protocol_uris_resolved

    @property
    def _protocol_uris_resolved(self):
        # FIXME quite slow ...
        for start_uri in self.protocol_uris:
            log.debug(start_uri)
            try:
                if not hasattr(start_uri, 'dereference'):
                    start_uri = idlib.StreamUri(start_uri)

                end_uri = start_uri.dereference()
                yield end_uri
                sc = end_uri.progenitor.status_code
                if sc > 400:
                    msg = f'error accessing {end_uri} {sc}'
                    if self.addError(msg, blame='submission'):
                        logd.error(msg)

            except idlib.exceptions.ResolutionError as e:
                pass  # FIXME I think we already log this error?
            except requests.exceptions.MissingSchema as e:
                if self.addError(e, blame='submission'):
                    logd.error(e)
            except OntId.BadCurieError as e:
                if self.addError(e, blame='submission'):
                    logd.error(e)
            except BaseException as e:
                #breakpoint()
                log.exception(e)
                log.critical('see exception above')

    @property
    def protocol_annotations(self):
        for uri in self.protocol_uris_resolved:
            yield from self.protc.byIri(uri, prefix=True)

    @property
    def protocol_jsons(self):
        for uri in self.protocol_uris_resolved:
            j = self._get_protocol_json(uri)
            if j:
                yield j

    def _get_protocol_json(self, uri):
        #juri = uri + '.json'
        logd.info(uri.identifier if isinstance(uri, idlib.Stream) else uri)  # FIXME
        pi = idlib.get_right_id(uri)
        if 'protocols.io' in pi:
            out = pi.data()
            if out is None and hasattr(pi, '_failure_message'):
                msg = pi._failure_message  # FIXME HACK use progenitor
                msg += ' {self.id!r}'
                if self.addError(msg):
                    logd.error(msg)

            return out
        else:
            msg = f'protocol uri is not from protocols.io {pi} {self.id}'
            if self.addError(msg):
                logd.error(msg)



import unittest
import pytest
from .common import project_path

from sparcur import schemas as sc
from sparcur import datasets as dat
from sparcur.paths import Path


class TestHierarchy(unittest.TestCase):
    def setUp(self):
        self.ds =  [dat.DatasetStructure(p) for p in Path(project_path).children]

    def tearDown(self):
        pass

    def test_create(self):
        ppattrs = project_path.cache.xattrs()
        for pthing in project_path.rglob('*'):
            if not pthing.skip_cache:
                ptattrs = pthing.cache.xattrs()

    def test_paths(self):
        for d in self.ds:
            for mp in d.meta_paths:
                print(mp)

        pytest.skip('TODO look at the lists here and figure out where they should go.')
        # for example if they are buried many levels too low how do we deal with that?

    def test_dataset(self):
        dsc = sc.DatasetStructureSchema()
        for d in self.ds:
            print(d.data)
            dsc.validate(d.data)

        pytest.skip('TODO look at the lists here and figure out where they should go.')

    def test_tables(self):
        for d in self.ds:
            for p in d.meta_paths:
                for row in dat.Tabular(p):
                    print(row)

    def test_submission(self):
        pass

    def test_dataset_description(self):
        pass

    def test_subjects(self):
        pass



import os
import atexit
import shutil
from tempfile import gettempdir
from pathlib import PurePosixPath
from datetime import datetime
import pytest
from augpathlib import PathMeta
from augpathlib.utils import onerror_windows_readwrite_remove
from sparcur import config
from sparcur import exceptions as exc
from sparcur.paths import Path
from sparcur.paths import LocalPath, PrimaryCache
from sparcur.paths import SymlinkCache
from sparcur.state import State
from sparcur.datasets import DatasetDescriptionFile
from sparcur.curation import PathData, Integrator
from sparcur.blackfynn_api import FakeBFLocal
this_file = Path(__file__)
examples_root = this_file.parent / 'examples'
template_root = this_file.parent.parent / 'resources/DatasetTemplate'
print(template_root)
_pid = os.getpid()
path_project_container = this_file.parent / f'test_local-{_pid}'
project_path = path_project_container / 'test_project'
fake_organization = 'N:organization:fake-organization-id'
project_path_real = path_project_container / 'UCSD'
test_organization = 'N:organization:ba06d66e-9b03-4e3d-95a8-649c30682d2d'
test_dataset = 'N:dataset:5d167ba6-b918-4f21-b23d-cdb124780da1'
temp_path = Path(gettempdir(), f'.sparcur-testing-base-{_pid}')

onerror = onerror_windows_readwrite_remove if os.name == 'nt' else None

SKIP_NETWORK = ('SKIP_NETWORK' in os.environ or
                'FEATURES' in os.environ and 'network-sandbox' in os.environ['FEATURES'])
skipif_no_net = pytest.mark.skipif(SKIP_NETWORK, reason='Skipping due to network requirement')

ddih = DatasetDescriptionFile.ignore_header  # save original skips
DatasetDescriptionFile.ignore_header = tuple(_ for _ in ddih if _ != 'example')  # use the example values for tests

ds_roots = (
    'ds1',
    'ds2/ds2',
    'ds3/oops',
    'ds3/too',
    'ds3/many',
    'ds4/really/deep',
    'ds5',
    'ds5/multi',
    'ds5/level/wat',
)
ds_folders = set(Path(d).parts[0] for d in ds_roots)

DIVISION_SLASH = '\u2215'  # what have we done


def mk_fldr_meta(fldr_path, ftype='collection', id=None):
    altid = f'N:{ftype}:' + fldr_path.as_posix()
    _id = id if id is not None else altid
    _id = _id.replace('/', DIVISION_SLASH)  # ARGH there must be a way to ensure file system safe ids :/
    _meta = fldr_path.meta
    kwargs = {**_meta}
    kwargs['id'] = _id
    meta = PathMeta(**kwargs)
    # NOTE st_mtime -> modified time of the file contents (data)
    # NOTE st_ctime -> changed time of the file status (metadata)
    # linux does not have a unified way to bet st_btime aka st_crtime which is birth time or created time
    return meta.as_xattrs(prefix='bf')
    #return {'bf.id': kwargs['id'],
            #'bf.created': meta.created.isoformat().replace('.', ',') if meta.created is not None else None,
            #'bf.updated': meta.updated.isoformat().replace('.', ',')}


def mk_file_meta(fp):
    meta = fp.meta
    return meta.as_xattrs(prefix='bf')
    #return {'bf.id': 'N:package:' + fp.as_posix(),
            #'bf.file_id': 0,
            #'bf.size': meta.size,
            # FIXME timezone
            #'bf.created': meta.created.isoformat().replace('.', ',') if meta.created is not None else None,
            #'bf.updated': meta.updated.isoformat().replace('.', ','),
            #'bf.checksum': fp.checksum(),
            # 'bf.old_id': None  # TODO
    #}


def mk_required_files(path, suffix='.csv'):
    # TODO samples.* ?!??!
    # FIXME because empty files get treated as non-existent
    # we either need to treat them as a different category of error
    # OR we need to fill in some fake values when testing
    for globpath in ('submission.*', 'dataset_description.*',  'subjects.*'):
        for template_file in template_root.glob(globpath):
            file_name = template_file.name
            file_path = path / file_name
            shutil.copy(template_file, file_path)
            file_path.touch()
            attrs = mk_file_meta(file_path)
            file_path.setxattrs(attrs)

if project_path.exists():
    # too much time wasted over stale test data
    # like ACSF just make it fresh every time
    project_path.rmtree(onerror=onerror)

project_path.mkdir(parents=True)
atexit.register(lambda : path_project_container.rmtree(onerror=onerror))
attrs = mk_fldr_meta(project_path, 'organization', id=fake_organization)
project_path.setxattrs(attrs)
# FIXME I know why lddi is on cache but I still don't like it
# because it should be possible to call lddi without forcing the cache to exist
# file under CachPath and RemotePath implementations are broken because instances
# can only exist if the underlying thing exists, which defeats the point of paths
project_path.cache.local_data_dir_init()

for ds in ds_folders:
    dsp = project_path / ds
    dsp.mkdir()
    dsp.setxattrs(mk_fldr_meta(dsp, 'dataset'))

for root in ds_roots:
    rp = project_path / root
    if not rp.exists():
        current_parent = rp
        to_reverse = []
        while not current_parent.exists():
            to_reverse.append(current_parent)
            current_parent = current_parent.parent

        for folder in reversed(to_reverse):
            folder.mkdir()
            folder.setxattrs(mk_fldr_meta(folder))

    print(rp)
    mk_required_files(rp)  # TODO all variants of missing files


fbfl = FakeBFLocal(project_path.cache.id, project_path.cache)
State.bind_blackfynn(fbfl)
#Integrator.setup()  # not needed for tests it seems



from sparcur import sheets
from sparcur import datasources as ds
#from sparcur import protocols

# state downstream of static external sources


class State:
    """ stateful values that many things need to access after startup
        that are beyond just the command line interface and which we
        don't want to continually create new versions of, in practice
        static information should flow from here rather than being set
        somewhere else and simply dumped here
    """

    @classmethod
    def bind_blackfynn(cls, blackfynn_local_instance):
        # FIXME bfli should flow from here out not the other way around
        # however there are some use cases, such as merging between
        # different organizations where you don't want to for the rest
        # of the program to be stuck with a single source, however for
        # our purposes here, we do need a way to say 'one at a time please'
        cls.blackfynn_local_instance = blackfynn_local_instance
        cls.member = ds.MembersData(blackfynn_local_instance)

    @classmethod
    def bind_protocol(cls, protocol_data):
        cls.protocol = protocol_data



import os
import secrets
import unittest
import augpathlib as aug
from sparcur import exceptions as exc
from sparcur.utils import GetTimeNow
from sparcur.paths import BlackfynnCache as BFC, LocalPath, Path
from sparcur.backends import BlackfynnRemote
from sparcur.blackfynn_api import BFLocal
from .common import test_organization, test_dataset, _pid
import pytest


class _TestOperation:
    @classmethod
    def tearDownClass(cls):
        base = aug.AugmentedPath(__file__).parent / f'test-operation-{_pid}'
        if base.exists():
            base.popd()  # in case we were inside it pop back out first
            base.rmtree()

    def setUp(self):
        class BlackfynnCache(BFC):
            pass

        BlackfynnCache._bind_flavours()

        base = aug.AugmentedPath(__file__).parent / f'test-operation-{_pid}'

        if base.exists():
            base.popd()  # in case we were inside it pop back out first
            base.rmtree()

        base.mkdir()
        base.pushd()

        self.BlackfynnRemote = BlackfynnRemote._new(LocalPath, BlackfynnCache)
        self.BlackfynnRemote.init(test_organization)
        self.anchor = self.BlackfynnRemote.dropAnchor(base)

        self.root = self.anchor.remote
        self.project_path = self.anchor.local
        list(self.root.children)  # populate datasets
        self.test_base = [p for p in self.project_path.children if p.cache.id == test_dataset][0]
        list(self.test_base.rchildren)  # populate test dataset

        asdf = self.root / 'lol' / 'lol' / 'lol (1)'

        class Fun(os.PathLike):
            name = 'hohohohohoho'

            def __fspath__(self):
                return ''

            @property
            def size(self):
                return aug.FileSize(len(b''.join(self.data)))

            @property
            def data(self):
                for i in range(100):
                    yield b'0' * 1000

        #wat = asdf.bfobject.upload(Fun(), use_agent=False)
        #breakpoint()


@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class TestDelete(_TestOperation, unittest.TestCase):

    def test_0(self):
        assert True

    def test_1_case(self):
        # this is an old scenario that happens because of how the old system worked
        # local working directory | x
        # local cache directory | o
        # remote | o
        pass


def make_rand(n, width=80):
    lw = width + 1
    hex_lw = lw // 2
    n_lines = n // lw
    n_accounted = n_lines * lw
    h_lines = n // (hex_lw * 2)
    h_accounted = h_lines * (hex_lw * 2)
    ldiff = n_lines - h_lines
    adiff = n_accounted - h_accounted
    accounted = n_accounted
    missing = n - accounted
    hex_missing = (missing + 1) // 2
    diff = hex_missing * 2 - missing
    hexstart = width % 2  # almost there fails on 71
    # also fails len(make_rand(102, 101)) - 102

    print(adiff, ldiff, missing, diff)
    string = '\n'.join([secrets.token_hex(hex_lw)[hexstart:]
                        for i in range(n_lines)]
                       + [secrets.token_hex(hex_missing)[diff:-1] + '\n'])
    return string.encode()


@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class TestUpdate(_TestOperation, unittest.TestCase):

    @pytest.mark.skip('the question has been answered')
    def test_process_filesize_limit(self):
        # so the 1 mb size file works, eventually, something else is wrong
        test_folder = self.test_base / 'hrm'
        test_folder.mkdir_remote()
        test_folder.__class__.upload = Path.upload
        for i in range(1024 ** 2, 5 * 1024 ** 2, 1024 ** 2):
            test_file = test_folder / f'size-{i}'
            if test_file.is_broken_symlink():
                test_file.remote.bfobject.package.delete()
                test_file.unlink()
                test_file = test_folder / f'size-{i}'  # remove stale cache

            test_file.data = iter((make_rand(i),))
            remote = test_file.upload()

    def test_upload_noreplace(self):
        for i in range(2):
            test_file = self.test_base / 'dataset_description.csv'
            test_file.data = iter((make_rand(100),))
            # FIXME temp sandboxing for upload until naming gets sorted
            test_file.__class__.upload = Path.upload
            # create some noise
            remote = test_file.upload(replace=False)
            print(remote.bfobject.package.name)

    def test_upload_noreplace_fail(self):
        # some persistent state from other tests is causing this to fail
        test_file = self.test_base / 'dataset_description.csv'
        test_file.data = iter((make_rand(100),))
        # FIXME temp sandboxing for upload until naming gets sorted
        test_file.__class__.upload = Path.upload
        test_file.upload(replace=False)
        try:
            test_file.upload(replace=False)
            assert False, 'should have failed'
        except exc.FileHasNotChangedError:
            pass

    def test_upload_replace(self):
        test_file = self.test_base / 'dataset_description.csv'
        test_file.data = iter((make_rand(100),))
        # FIXME temp sandboxing for upload until naming gets sorted
        test_file.__class__.upload = Path.upload
        test_file.upload()


@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class TestClone(_TestOperation, unittest.TestCase):
    # TODO test a variety of clone scenarios
    # and consider whether testing for and
    # existing root should be done in dropAnchor

    def setUp(self):
        super().setUp()
        self.alt_project_path = self.project_path.parent / 'alt' / self.project_path.name
        if self.alt_project_path.parent.exists():
            self.alt_project_path.parent.rmtree()

        self.alt_project_path.mkdir(parents=True)

    def _do_target(self, target, expect_error_type=None):
        class BlackfynnCache(BFC):
            pass

        BlackfynnCache._bind_flavours()

        BFR = BlackfynnRemote._new(LocalPath, BlackfynnCache)
        BFR.init(test_organization)

        if expect_error_type:
            try:
                anchor = BFR.dropAnchor(target.parent)
                raise AssertionError(f'should have failed with a {expect_error_type}')
            except expect_error_type as e:
                pass

        else:
            anchor = BFR.dropAnchor(target.parent)

    def test_1_in_project(self):
        target = self.project_path / 'some-new-folder'
        target.mkdir()
        self._do_target(target)  # FIXME succeeds for now, but probably should not?

    def test_2_project_top_level(self):
        target = self.project_path
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_3_existing_empty(self):
        target = self.alt_project_path
        self._do_target(target)

    def test_4_existing_has_folder(self):
        target = self.alt_project_path
        child = target / 'a-folder'
        child.mkdir(parents=True)
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_5_existing_has_file(self):
        target = self.alt_project_path
        child = target / 'a-file'
        child.touch()
        self._do_target(target, exc.DirectoryNotEmptyError)

    def test_6_existing_has_local_data_dir(self):
        target = self.alt_project_path
        child = target / self.anchor._local_data_dir
        child.mkdir()
        self._do_target(target, exc.DirectoryNotEmptyError)


class TestMkdirRemote(_TestOperation, unittest.TestCase):

    def test_mkdir_remote_parents_false(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP}' / 'some-folder'
        try:
            remote = local.mkdir_remote()
            raise AssertionError('Should have failed since parents=False')
        except FileNotFoundError:
            pass

    def test_0_mkdir_remote_will_be_dataset(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP}'
        remote = local.mkdir_remote()
        remote.rmdir()
        remote.cache.refresh()  # reminder that remotes are a snapshot in time, NOT dynamic
        assert not local.exists(), f'should have been deleted {remote}'

    def test_1_mkdir_remote_will_be_collection(self):
        now = GetTimeNow()
        local = self.project_path / f'test-dataset-{now.START_TIMESTAMP}' / 'some-folder'
        remote = local.mkdir_remote(parents=True)
        parent = remote.parent
        try:
            parent.rmdir()  # should fail here
            try:
                remote.rmdir()  # insurance
            except BaseException as e:
                log.exception(e)
            finally:
                raise AssertionError('remote parent should NOT have rmdired {parent}')
        except exc.PathNotEmptyError:
            pass

        try:
            remote.rmdir()
            remote.cache.refresh()
            assert not local.exists(), f'should have been deleted {remote}'
        finally:
            lparent = parent.local
            parent.cache.refresh()  # we just removed the child so the parent is stale
            parent.rmdir()
            parent.cache.refresh()  # and THIS is the error we have been trying to handle all night!
            assert not lparent.exists(), f'should have been deleted {parent}'


class TestRemote(_TestOperation, unittest.TestCase):
    def test_remote_path_does_not_exist(self):
        new_thing = self.root / 'does not exist'

    @pytest.mark.skip('Not ready.')
    def test_cache_path_does_not_exist(self):
        """ This should not produce an error.
            Path objects should be able to be instantiated without
            po.exists() -> True at a point in time prior to instantiation.
        """
        new_thing = self.anchor / 'does not exist'

    def __test_cache_path_fake_id(self):
        # FIXME the right way to do this is
        np = self.project_path / 'new-path'
        np.mkdir()
        npm = np.meta
        
        # bad way
        class FakeBase(aug.RemotePath):
            def __init__(self, id, name, cache=None):
                super().__init__(id, cache)
                self.name = name
                now = GetTimeNow()
                self.created = now._start_time
                self.updated = now._start_time
                self.checksum = 'lolnone'
                self.chunksize = 4096
                self.file_id = 'asdfasdfasdf'

            @property
            def meta(self):
                return PathMeta(size=self.size,
                                created=self.created,
                                updated=self.updated,
                                checksum=self.checksum,
                                chunksize=self.chunksize,
                                id=self.id,
                                file_id=self.file_id)


            @property
            def parent(self):
                return None

            def _parts_relative_to(self, remote, cache_parent=None):
                return [self.name]


        # This takes way too many steps :/
        Fake = FakeBase._new(LocalPath, aug.CachePath)

        fake = Fake('lol', 'double lol')

        self.anchor / fake





import io
import csv
import copy
from types import GeneratorType
from itertools import chain
from collections import Counter
import augpathlib as aug
from xlsx2csv import Xlsx2csv, SheetNotFoundException, InvalidXlsxFileException
from terminaltables import AsciiTable
from pyontutils.utils import byCol, Async, deferred, python_identifier
from pyontutils.namespaces import OntCuries, makeNamespaces, TEMP, isAbout
from pyontutils.namespaces import rdf, rdfs, owl, skos, dc
from . import schemas as sc
from . import raw_json as rj
from . import exceptions as exc
from . import normalization as nml
from .core import log, logd, HasErrors
from .paths import Path
from .utils import is_list_or_tuple

a = rdf.type


# FIXME review this :/ it is not a good implementation at all
def tos(hrm):
    if isinstance(hrm, GeneratorType):
        t = tuple(hrm)
        if not t:
            return
        elif len(t) == 1:
            return t[0]
        else:
            return t
    else:
        return hrm


hasSchema = sc.HasSchema()
@hasSchema.mark
class Header:
    """ generic header normalization for python """
    # FIXME this is a really a pipeline stage
    def __init__(self, first_row_or_column, normalize=True):
        self.pipeline_start = first_row_or_column
        self._normalize = normalize

    @property
    def normalized(self):
        orig_header = self.pipeline_start
        header = []
        for i, c in enumerate(orig_header):
            if c:
                c = python_identifier(c)
                c = nml.NormHeader(c)

            if not c:
                c = f'TEMP_{i}'

            if c in header:
                c = c + f'_TEMP_{i}'

            header.append(c)

        return header

    @hasSchema(sc.HeaderSchema)
    def data(self):
        return self.normalized

    @property
    def _lut(self):
        return {n:h for n, h in zip(self.data, self.pipeline_start)}

    def lut(self, **renames):
        if self._normalize:
            return self._lut_normalize(renames)
        else:
            return self._lut_original(renames)

    def _lut_normalize(self, renames):
        """ do any renames beyond the specific normalization """
        return {renames[n] if n in renames else n:h
                for n, h in zip(self.data, self.pipeline_start)}

    def _lut_original(self, renames):
        original = self.pipeline_start
        normalized = self.normalized
        if len(set(original)) != len(original):
            dupes = [v for v, c in Counter(original).most_common() if c > 1]
            msg = f'Original header is not unique.\nDuplicate entries: {dupes}'
            raise exc.MalformedHeaderError(msg)

        return {renames[n] if n in renames else n:h
                for n, h in zip(original, original)}


class DatasetMetadata(Path, HasErrors):

    @property
    def data(self):
        if self.cache is not None:
            cmeta = self.cache.meta
            return dict(
                id=self.cache.id,
                meta=dict(
                    folder_name=self.name,
                    uri_human=self.cache.uri_human,
                    uri_api=self.cache.uri_api,
                    timestamp_created=cmeta.created,
                    timestamp_updated=cmeta.updated,
                    timestamp_updated_contents=self.updated_cache_transitive(),
                        ))
        else:
            return dict(id=self.id,
                        meta=dict(folder_name=self.name,
                                  uri_human=None,
                                  uri_api=None,
                                  timestamp_created=None,
                                  timestamp_updated=None,
                        ))


DatasetMetadata._bind_flavours()


class DatasetStructure(Path):
    sections = None  # NOTE assigned in at end of file
    rglobs = 'manifest',
    default_glob = 'glob'
    max_childs = 40
    rate = 8  # set by Integrator from cli
    _refresh_on_missing = True

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        DatasetStructure._setup(*args, **kwargs)
        DatasetStructure.__new__ = DatasetStructure._renew
        return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def _setup(*args, **kwargs):
        import requests
        DatasetStructure._requests = requests

    @property
    def children(self):
        # for whatever reason the way pathlib constructs children
        # skips calling init, iirc because they use object.__new__
        # or something

        for child in super().children:
            child.__init__()
            yield child

    @classmethod
    def _bind_sections(cls):
        for section, _sec_class in cls.sections.items():
            @property
            def sec(self, sec_class=_sec_class, name='_cache_' + _sec_class.__name__):
                if not hasattr(self, name):
                    mp = [p for p in self.meta_paths if isinstance(p, sec_class)]
                    if len(mp) == 1:
                        setattr(self, name, mp[0])
                    elif not mp:
                        setattr(self, name, None)
                    else:
                        setattr(self, name, mp)

                return getattr(self, name)

            setattr(cls, section, sec)

    def files_last_updated(self):
        # FIXME TODO can we get this more efficiently during pull?
        all_updated = [c.cache.meta.updated for c in
                       chain((self,), self.rchildren) if c.is_file()]
        if all_updated:
            return max(all_updated)

    @property
    def counts(self):
        if not hasattr(self, '_counts'):
            size = 0
            dirs = 0
            files = 0
            need_meta = []
            if not self.is_dir():
                gen = self,

            else:
                gen = self.rchildren

            for c in gen:
                if c.is_dir():
                    dirs += 1
                    maybe_size = 0  # the remote size of a folder is zero for bf
                else:
                    files += 1  # testing for broken symlinks is hard
                    try:
                        if c.is_broken_symlink():
                            maybe_size = (aug.PathMeta
                                          .from_symlink(c)
                                          .size)
                        else:
                            maybe_size = (aug.PathMeta
                                          .from_xattrs(c.xattrs(),
                                                        prefix='bf')
                                          .size)

                        #maybe_size = c.cache.meta.size
                    except AttributeError as e:
                        log.error(f'no cache or no meta for {c}\n{e}')
                        continue

                    if maybe_size is None:
                        need_meta.append(c)
                    else:
                        size += maybe_size

            if need_meta and self._refresh_on_missing:
                log.critical(f'OH NO THIS IS GOING TO CAUSE DUPLICATES (probably)\n{self}')
                nl = '\n'
                log.info(f'refreshing {len(need_meta)} files with missing metadata in {self}'
                         f'\n{nl.join(_.as_posix() for _ in need_meta)}')
                breakpoint()
                raise TypeError('fuck')
                new_caches = Async(rate=self.rate)(deferred(c.cache.refresh)() for c in need_meta)
                for c in new_caches:  # FIXME first time around meta doesn't get updated ??
                    if c is None:
                        continue  # file was deleted (logged previously)

                    if c.meta is None:
                        log.critical(f'missing metdata! {c}')
                        continue

                    size += c.meta.size

            self._counts = dict(size=aug.FileSize(size), dirs=dirs, files=files)

        return self._counts

    @property
    def total_size(self):
        """ total remote size, trigger retrieval of all remote metadata """
        return self.counts['size']

    @property
    def total_dirs(self):
        return self.counts['dirs']

    @property
    def total_files(self):
        return self.counts['files']

    @property
    def total_paths(self):
        return self.counts['dirs'] + self.counts['files']

    @property
    def bids_root(self):
        # FIXME this will find the first dataset description file at any depth!
        # this is incorrect behavior!
        """ Sometimes there is an intervening folder. """
        if self.cache.is_dataset:
            def check_fordd(paths, level=0, stop=3):
                if not paths:  # apparently the empty case recurses forever
                    return

                if len(paths) > self.max_childs:
                    log.warning(f'Not globing in a folder with > {self.max_childs} children! '
                                f'{self.as_posix()!r}')
                    return
                dd_paths_all = []
                children = []
                for path in paths:
                    dd_paths = list(path.glob('[Dd]ataset_description*.*'))
                    if dd_paths:
                        dd_paths_all.extend(dd_paths)
                    elif not dd_paths_all:
                        children.extend([p for p in path.children if p.is_dir()])

                if dd_paths_all:
                    return dd_paths_all
                else:
                    return check_fordd(children, level + 1)

            dd_paths = check_fordd((self,))

            if not dd_paths:
                #log.warning(f'No bids root for {self.name} {self.id}')  # logging in a property -> logspam
                return

            elif len(dd_paths) > 1:
                #log.warning(f'More than one submission for {self.name} {self.id} {dd_paths}')
                pass

            return dd_paths[0].parent  # FIXME choose shorter version? if there are multiple?

        elif self.parent:  # organization has no parent
            return self.parent.bids_root

    def _abstracted_paths(self, name_prefix, glob_type=None):
        """ A bottom up search for the closest file in the parent directory.
            For datasets, if the bids root and path do not match, use the bids root.
            In the future this needs to be normalized because the extra code required
            for dealing with the intervening node is quite annoying to maintain.
        """
        if glob_type is None:
            glob_type = self.default_glob

        path = self
        if (self.cache and
            self.cache.is_dataset and
            self.bids_root is not None and
            self.bids_root != self):
            path = self.bids_root

        first = name_prefix[0]
        cased_np = '[' + first.upper() + first + ']' + name_prefix[1:]  # FIXME warn and normalize
        glob = getattr(path, glob_type)
        gen = glob(cased_np + '*.*')

        try:
            path_class = self.sections[name_prefix]
            path = next(gen)
            for path in chain((path,), gen):
                path = path_class(path)
                if path.is_broken_symlink():
                    log.info(f'fetching unretrieved metadata path {path.as_posix()!r}'
                             '\nFIXME batch these using async in cli export ...')
                    try:
                        path.cache.fetch(size_limit_mb=path.cache.meta.size.mb + 1)
                    except self._requests.exceptions.ConnectionError as e:
                        raise exc.NetworkFailedForPathError(path) from e

                if path.suffix in path.stem:
                    msg = f'path has duplicate suffix {path.as_posix()!r}'
                    self.addError(msg,
                                  blame='submission',
                                  path=path)
                    logd.error(msg)

                if path.name[0].isupper():
                    msg = f'path has bad casing {path.as_posix()!r}'
                    if self.addError(msg,
                                     blame='submission',
                                     path=path):
                        logd.error(msg)

                yield path

        except StopIteration:
            if (self.cache.parent.meta is not None and
                self.parent.cache != self.cache.anchor and
                self.parent != self):
                yield from getattr(self.parent, name_prefix + '_paths')

    @property
    def meta_paths(self):
        for section_name in self.sections:
            glob_type = 'rglob' if section_name in self.rglobs else None
            yield from self._abstracted_paths(section_name, glob_type=glob_type)

    @property
    def data(self):
        out = self.counts
        for section_name in self.sections:
            #section_paths_name = section_name + '_paths'
            #paths = list(getattr(self, section_paths_name))
            glob_type = 'rglob' if section_name in self.rglobs else 'glob'
            paths = list(self._abstracted_paths(section_name, glob_type=glob_type))
            section_key = section_name + '_file'  # XXX _file is added here
            if paths:
                if len(paths) == 1 and section_name != 'manifest':  # FIXME manifest is a hack
                    path, = paths
                    out[section_key] = path
                else:
                    out[section_key] = paths

        self.embedErrors(out)
        return out


class ObjectPath(Path):
    obj = None
    @property
    def object(self):
        return self.obj(self)


ObjectPath._bind_flavours()


class Tabular(HasErrors):

    def __init__(self, path, rotate=False):
        super().__init__()
        self.path = path
        self._rotate = rotate

    @property
    def file_extension(self):
        if self.path.suffixes:
            ext = self.path.suffixes[0]  # FIXME filenames with dots in them ...
            if ext != '.fake':
                return nml.NormFileSuffix(ext).strip('.')

    def tsv(self):
        return self.csv(delimiter='\t')

    def csv(self, delimiter=','):
        # FIXME if the number of delimiters in a row other than the header
        # is greater than the number of delimiters in the header then any
        # columns beyond the number in the header will be truncated and
        # I'm not sure that this loading routine detects that
        for encoding in ('utf-8', 'latin-1'):
            try:
                with open(self.path, 'rt', encoding=encoding) as f:
                    rows_orig = list(csv.reader(f, delimiter=delimiter))

                rows = [row for row in rows_orig if row and any(row)]
                diff = len(rows_orig) - len(rows)
                if diff:
                    # LOL THE FILE WITH > 1 million empty rows, 8mb of commas
                    # totally the maximum errors champion
                    message = f'There are {diff} empty rows in {self.path.as_posix()!r}'
                    if self.addError(message,
                                     blame='submission',
                                     path=self.path):
                        logd.error(message)

                if encoding != 'utf-8':
                    message = f'encoding bad {encoding!r} {self.path.as_posix()!r}'
                    if self.addError(exc.EncodingError(message),
                                     blame='submission',
                                     path=self.path):
                        logd.error(message)

                yield from rows
                return
            except UnicodeDecodeError:
                continue
            except csv.Error as e:
                logd.exception(e)
                message = f'WHAT HAVE YOU DONE {e!r} {self.path.as_posix()!r}'
                if self.addError(message,
                                 blame='EVERYONE',
                                 path=self.path):
                    logd.error(message)

    def xlsx(self):
        kwargs = {
            'delimiter': '\t',
            'skip_empty_lines': True,
            'outputencoding': 'utf-8',
            'hyperlinks': True,
        }
        sheetid = 1
        try:
            xlsx2csv = Xlsx2csv(self.path.as_posix(), **kwargs)
        except InvalidXlsxFileException as e:
            raise exc.NoDataError(f'{self.path}') from e

        ns = len(xlsx2csv.workbook.sheets)
        if ns > 1:
            message = f'too many sheets ({ns}) in {self.path.as_posix()!r}'
            if self.addError(exc.EncodingError(message),
                             blame='submission',
                             path=self.path):
                logd.error(message)

        f = io.StringIO()
        try:
            xlsx2csv.convert(f, sheetid)
            f.seek(0)
            gen = csv.reader(f, delimiter='\t')
            yield from gen
        except SheetNotFoundException as e:
            log.warning(f'Sheet weirdness in {self.path}')
            log.warning(str(e))
        except AttributeError as e:
            message = ('Major sheet weirdness (maybe try resaving, '
                       'probably a bug in the xlsx2csv converter)? '
                       f'in {self.path}')
            if self.addError(exc.EncodingError(message),
                             blame='submission',
                             path=self.path):
                log.exception(e)
                logd.critical(message)

    def _bad_filetype(self, type_):
        message = f'bad filetype {type_}\n{self.path.as_posix()!r}'
        raise exc.FileTypeError(message)

    def xls(self):
        return self._bad_filetype('xls')

    def normalize(self, rows):
        # FIXME need to detect changes
        # this removes any columns that are all dead

        #if any(not(c) for c in rows[0]):  # doesn't work because generators

        error = exc.EncodingError(f"encoding feff error in '{self.path}'")
        cleaned_rows = zip(*(t for t in zip(*rows) if not all(not(e) for e in t)))  # TODO check perf here
        for row in cleaned_rows:
            n_row = [c.strip().replace('\ufeff', '') for c in row
                     if ((not logd.error(error) if self.addError(error) else True)  # FIXME will probably append multiple ...
                         if '\ufeff' in c else True)]
            if not all(not(c) for c in n_row):  # skip totally empty rows
                yield n_row

    @property
    def normalized(self):
        try:
            yield from self.normalize(getattr(self, self.file_extension)())
        except UnicodeDecodeError as e:
            log.error(f'{self.path.as_posix()!r} {e}')

    @property
    def T(self):
        return self.__class__(self.path, rotate=True)

    def __iter__(self):
        yield from self._iter_(normalize=True)

    def _iter_(self, normalize=False):
        try:
            if normalize:
                try:
                    fef = getattr(self, self.file_extension)
                except AttributeError as e:
                    self._bad_filetype(self.file_extension)

                gen = self.normalize(fef())
            else:
                gen = getattr(self, self.file_extension)()

            if self._rotate:
                yield from zip(*gen)
            else:
                yield from gen
        except UnicodeDecodeError as e:
            log.error(f'{self.path.as_posix()!r} {e}')

    @property
    def title(self):
        path = Path(self.path)
        out = path.name
        if path.cache is not None and path.cache.dataset is not None:
            out += ' ' + path.cache.dataset.name[:30] + ' ...'

        return out

    def __repr__(self):
        limit = 30
        rows = [[c[:limit] + ' ...' if isinstance(c, str)
                 and len(c) > limit else c
                 for c in r] for r in self]
        table = AsciiTable(rows, title=self.title)
        return table.table


ROW_TYPE = object()
COLUMN_TYPE = object()
numberN = object()
GROUP_ALL = object()


def remove_rule(blob, rule):
    # FIXME this should be in _DictTransformer
    previous_target = blob
    target = blob
    previous_element = None
    for path_element in rule:
        if isinstance(path_element, str):
            if path_element in target:
                previous_target = target
                target = previous_target[path_element]
            else:
                return  # target not found

        elif isinstance(path_element, dict):
            if all([(k in target and target[k] == v)
                    for k, v in path_element.items()]):
                previous_target.pop(previous_element)
                return  # we're done here

        else:
            raise ValueError(f'what are you doing with {rule}')

        previous_element = path_element


class PrimaryKey:
    def __init__(self, name, column_names, combine_function,
                 nalt_header, agen, gen):
        self.name = name
        # because we add the primary key to alt header we remove it again here
        # because the rest of the generator values don't have a blank prepended
        nalt_header_less_pk = [nah for nah in nalt_header if nah != name]
        try:
            self.indexes = [nalt_header_less_pk.index(name) for name in column_names]
        except ValueError as e:
            msg = f'Missing a primary key component {e}'
            raise exc.MalformedHeaderError(msg) from e

        self.combine_function = combine_function
        self._agen = agen
        self._gen = gen

    def compute_primary(self, arecord):
        return self.combine_function(tuple(arecord[index] for index in self.indexes))

    def __iter__(self):
        yield (self.name, *[self.compute_primary(ar) for ar in self._agen])
        yield from self._gen

    @property
    def generator(self):
        return iter(self)


class MetadataFile(HasErrors):
    default_record_type = ROW_TYPE
    primary_key_rule = None  # name, the names of the alt_header columns to merge, function to merge tuple
    renames_alt = {}     # renames first since all following
    renames_header = {}  # refs look for the renamed form
    missing_add = {}  # missing also comes after renames
    record_type_key_alt = None  # record type key is affected by renames
    record_type_key_header = None  # record type key is affected by renames
    groups_alt = {}
    groups_header = {}
    ignore_alt = tuple()
    ignore_header = tuple()
    ignore_match = tuple()
    raw_json_class = rj.RawJson
    normalization_class = nml.NormValues
    normalize_alt = True
    normalize_header = True
    _expect_single = tuple()

    def __new__(cls, path, schema_version=None):
        if schema_version is not None:  # and schema_version < latest  # TODO
            logd.info(schema_version)  # TODO warn

        if cls.record_type_key_header is None or cls.record_type_key_alt is None:
            raise TypeError(f'record_type_key_? should not be None on {cls.__name__}')

        # TODO check for renames that didn't get picked up
        # especially from the 0, 0 cell

        return super().__new__(cls)

    def __init__(self, path, schema_version=None):
        super().__init__()
        self.path = path
        self.schema_version = schema_version

    def xopen(self):
        self.path.xopen()

    @property
    def data(self):
        data = self._data()

        condition = False
        if condition:
            breakpoint()

        return data

    def _data(self):
        if self.path.suffix == '.json':
            data = self.raw_json_class(self.path).data
            self._expand_string_lists = lambda : data  # FIXME hack
            self.norm_to_orig_header = {}
            def all_keys(d):
                if isinstance(d, dict):
                    for k, v in d.items():
                        if k not in self.groups_alt:
                            yield k

                        yield from all_keys(v)

                elif isinstance(d, list):
                    for e in d:
                        yield from all_keys(e)

            self.norm_to_orig_alt = {k:k for k in all_keys(data)}
            #return self.normalization_class(self).data  # FIXME incredible violation of encapsulation ...
            return self._condense()

        else:
            # TODO if this passes we can just render backward from _normalize
            return self._condense()

    def _condense(self):
        def condense(thing, key=None):
            if isinstance(thing, dict):
                hrm = []
                out = {}
                for k, v in thing.items():
                    nv = condense(v, k)
                    # bah bloody description appearing twice >_<
                    if k in self.norm_to_orig_header and k not in self.norm_to_orig_alt:
                        if is_list_or_tuple(nv):
                            # at this stage there are no lists of lists
                            hrm.extend(nv)
                        else:
                            hrm.append(nv)
                    else:
                        out[k] = nv

                if out and hrm:
                    raise ValueError(f'how!?\n{hrm}\n{out}')
                elif out:
                    return out
                elif hrm:
                    if (key in self._expect_single and
                        len(hrm) == 1):
                        return hrm[0]
                    else:
                        return tuple(hrm)

            #elif is_list_or_tuple(thing):
                #return thing
            else:
                return thing

        data_in = self._cull_rtk()
        data_out = condense(data_in)
        if data_out is None:
            # None breaks further pipelines
            # yes types would help here
            return {}
        else:
            return data_out

    def _cull_rtk(self):
        normalized = self._normalize_values()
        def crtk(thing, only_index_alt=False, only_index_header=False):
            if isinstance(thing, dict):
                out = {}
                for k, v in thing.items():
                    if k in self.groups_alt and self.groups_alt[k] == GROUP_ALL:
                        nv = crtk(v, only_index_alt=True)
                        if nv is not None:
                            out[k] = nv
                    elif k in self.groups_header and self.groups_header[k] == GROUP_ALL:
                        # this isn't used yet but putting it here for symmetry just in case
                        nv = crtk(v, only_index_header=True)
                        if nv is not None:
                            out[k] = nv
                    elif (only_index_alt and k == v == self.record_type_key_alt or
                          only_index_header and k == v == self.record_type_key_header):
                        continue
                    elif (k != self.record_type_key_header and
                          # FIXME requires that headers and alts be fully disjoint
                          k != self.record_type_key_alt or
                          only_index_alt and k == self.record_type_key_alt or
                          only_index_header and k == self.record_type_key_header):
                        nv = crtk(v, only_index_alt, only_index_header)
                        if nv is not None:
                            out[k] = nv

                if out:
                    return out

            else:
                return thing

        crv = crtk(normalized)
        return crv

    def _normalize_values(self):
        nc = self.normalization_class(self)  # calls _normalize_keys internally
        return nc.data

    def _clean(self):
        data_in = self._expand_string_lists()
        def clean(thing):
            if isinstance(thing, dict):
                out = {}
                for k, v in thing.items():
                    cv = clean(v)
                    if cv is not None:
                        out[k] = cv

                if out:
                    return out

            elif thing == '' or thing == {}:
                return None

            elif is_list_or_tuple(thing):
                return [v for v in [clean(v) for v in thing] if v is not None]

            else:
                return thing

        data_out = clean(data_in)
        return data_out

    def _expand_string_lists(self):
        data_in = self._filter()
        # TODO consider whether this step would be useful
        # I think it would because it would vastly simplify
        # the downstream implementation of things like
        # normalizing the contributor roles
        # TODO this might also go before filter?
        # might need to blacklist or whitelist fields?
        seps = '|', ';', ','
        data_out = data_in
        return data_out

    def _filter(self):
        # FIXME does in place modification
        culled = self._cull()
        for rule in self.ignore_match:
            remove_rule(culled, rule)

        return culled

    def _cull(self):
        transformed = self._transformed()
        # FIXME deeper nesting might break this?
        culled = {k:{k:v for k, v in v.items()
                     if k not in self.ignore_header}
                  for k, v in transformed.items()}
        return culled

    def _transformed(self):
        keyed = self._keyed()
        naccounted_baseline = set(e for g in self.groups_alt.values()
                                  if g != GROUP_ALL for e in g)

        # TODO how to flag exclude the header fields ?
        transformed = {}
        for key, alt_grouped_norm in self.groups_alt.items():
            merge = alt_grouped_norm == GROUP_ALL
            if merge:
                alt_grouped_norm = tuple(k for k in self.norm_to_orig_alt
                                         if k not in naccounted_baseline)
                naccounted_baseline.update(set(alt_grouped_norm))

            records = [keyed[nh] for nh in alt_grouped_norm
                       # FIXME versioning ... schema will take care of missing values later
                       if nh in keyed]

            objected = {nh:{record[self.record_type_key_header]:record[nh]
                            for record in records}
                        for nh in self.norm_to_orig_header}

            transformed[key] = {k:v for k, v in objected.items()
                                if v is not None and v != {}}

        nunaccounted = (set(self.norm_to_orig_alt) - naccounted_baseline)

        for nah in nunaccounted: 
            transformed[nah] = keyed[nah]

        return transformed

    def _keyed(self):
        objects = self._add_missing()
        keyed = {o[self.record_type_key_header]:o for o in objects}
        return keyed

    def _add_missing(self):
        objects = self._sanity_check()
        for nah, (value, number) in self.missing_add.items():
            if nah not in self.norm_to_orig_alt:
                self.norm_to_orig_alt[nah] = nah
                cant_go_wrong = {}
                i = 0
                for nh in self.norm_to_orig_header:
                    if ((number == numberN or i < number) and
                        nh != self.record_type_key_header and
                        nh not in self.ignore_header):
                        v = value
                        i += 1
                    else:
                        v = ''

                    cant_go_wrong[nh] = v

                {nh:(value if number == numberN or i <= number else '')
                 for i, nh in enumerate(self.norm_to_orig_header)
                 if nh != self.record_type_key_header}
                #print(cant_go_wrong)
                cant_go_wrong[self.record_type_key_header] = nah
                objects.append(cant_go_wrong)

        # end add missing items
        return objects

    def _sanity_check(self):
        """ first chance we have to check sanity since
            the generators have to have run """
        objects = self._objectify()
        for n, rtk, norm_dict in (('alt_header', self.record_type_key_alt, self.norm_to_orig_alt),
                                  ('header', self.record_type_key_header, self.norm_to_orig_header)):
            if rtk not in norm_dict:
                #breakpoint()  # XXX TODO enable this on debug
                msg = (f'Could not find record primary key for {n} in\n"{self.path}"\n\n'
                       f'{rtk}\nnot in\n{list(norm_dict)}')

                raise exc.MalformedHeaderError(msg)

        return objects

    def _objectify(self):
        gen = self._norm_alt_headers()
        # WARNING assumes dict order
        objects = [{k:v for k, v in zip(self.norm_to_orig_header, r)} for r in gen]
        return objects

    def _norm_alt_headers(self):
        # headers MUST be normalized as early as possible due to the fact that
        # there are very likely to be duplicate keys
        gen = self._headers()

        def normb(k):
            if k in self.orig_to_norm_alt:
                nk = self.orig_to_norm_alt[k]
            else:
                nk = k

            return nk

        for row in gen:
            yield (normb(row[0]), *row[1:])

    def _headers(self):
        t = self._t()

        if self.default_record_type == ROW_TYPE:
            alt = t
            head = t.T
        else:
            alt = t.T
            head = t

        agen = iter(alt)
        gen = iter(head)

        try:
            if self.primary_key_rule is not None:
                pk_name, combine_names, combine_function = self.primary_key_rule
                self.alt_header = (pk_name, *next(agen))
            else:
                self.alt_header = next(agen)
        except StopIteration as e:
            raise exc.NoDataError(f'{self.path}') from e

        self._alt_header = Header(self.alt_header, normalize=self.normalize_alt)
        self.norm_to_orig_alt = self._alt_header.lut(**self.renames_alt)
        self.orig_to_norm_alt = {v:k for k, v in self.norm_to_orig_alt.items()}

        if self.primary_key_rule is not None:
            nalt_header = [self.orig_to_norm_alt[ah] for ah in self.alt_header]
            gen = PrimaryKey(pk_name, combine_names, combine_function,
                             nalt_header, agen, gen).generator

        # FIXME if the 'header' column is not at position zero
        # this will fail see version 1.1 subjects template for this
        # with a note that 1.1 had many issues
        self.header = next(gen)
        self._header = Header(self.header, normalize=self.normalize_header)
        self.norm_to_orig_header = self._header.lut(**self.renames_header)
        self.orig_to_norm_header = {v:k for k, v in self.norm_to_orig_header.items()}

        yield self.header
        yield from gen

    def _t(self):
        if self.path.suffix == '.json':
            return  # TODO json -> tabular conversion
        else:
            return Tabular(self.path)


class SubmissionFile(MetadataFile):
    __internal_id_1 = object()
    default_record_type = COLUMN_TYPE
    renames_alt = {'submission_item': __internal_id_1}
    renames_header = {'definition': 'definition_header'}
    record_type_key_alt = __internal_id_1
    record_type_key_header = 'submission_item'
    groups_alt = {'submission': ('sparc_award_number', 'milestone_achieved', 'milestone_completion_date')}
    ignore_alt = __internal_id_1,
    ignore_header = 'definition_header',
    raw_json_class = rj.RawJsonSubmission
    normalization_class = nml.NormSubmissionFile

    @property
    def data(self):
        """ lift list with single element to object """

        data = super().data  # LOL PYTHON can't super in a debugger -- pointless
        d = copy.deepcopy(data)

        if d and 'submission' in d:
            sub = d['submission']
            if sub:
                if is_list_or_tuple(sub):
                    d['submission'] = sub[0]
                elif isinstance(sub, dict):
                    return d
                else:
                    raise TypeError(f'What is a {type(sub)} {sub}?')
            else:
                d['submission'] = {}

        return d


class SubmissionFilePath(ObjectPath):
    obj = SubmissionFile


SubmissionFilePath._bind_flavours()


_props = sc.DatasetDescriptionSchema.schema['properties']
_props2 = sc.ContributorSchema.schema['properties']  # FIXME recurse ...
_nddfes = [k for k, v in chain(_props.items(), _props2.items())
           if isinstance(v, dict) and sc.not_array(v)]
_nddfes = sorted(set(_nddfes))


class DatasetDescriptionFile(MetadataFile):
    default_record_type = COLUMN_TYPE
    renames_alt = {'contributors': 'contributor_name',  # watch out for name collisions (heu heu heu)
                   'metadata_version_do_not_change': 'schema_version'}
    renames_header = {'description': 'description_header'}
    missing_add = {'schema_version': ('1.0', 1)}
    record_type_key_alt = 'metadata_element'
    record_type_key_header = record_type_key_alt
    groups_alt = {'contributors': ('contributor_name',
                                   'contributor_orcid_id',
                                   'contributor_affiliation',
                                   'contributor_role',
                                   'is_contact_person',),
                  'links': ('additional_links', 'link_description'),
                  'examples': ('example_image_filename',
                               'example_image_locator',
                               'example_image_description'),}

    ignore_header = 'metadata_element', 'example', 'description_header'
    raw_json_class = rj.RawJsonDatasetDescription
    normalization_class = nml.NormDatasetDescriptionFile
    _expect_single = _nddfes

    @property
    def data(self):
        data = super().data
        #breakpoint()
        return data


class DatasetDescriptionFilePath(ObjectPath):
    obj = DatasetDescriptionFile


DatasetDescriptionFilePath._bind_flavours()


_props = sc.SubjectsSchema.schema['properties']['subjects']['items']['properties']
_props2 = sc.SamplesFileSchema.schema['properties']['samples']['items']['properties']
_props3 = sc._software_schema['items']['properties']
_nsffes = [k for k, v in chain(_props.items(), _props2.items(), _props3.items())
           if isinstance(v, dict) and sc.not_array(v)]
_nsffes = sorted(set(_nsffes))


class SubjectsFile(MetadataFile):
    #default_record_type = COLUMN_TYPE
    __internal_id_1 = object()
    renames_header = {'subject_id': 'metadata_element',
                      ('Lab-based schema for identifying each subject, '
                       'should match folder names'): __internal_id_1,}
    record_type_key_alt = 'subject_id'
    record_type_key_header = 'metadata_element'
    groups_alt = {'subjects': GROUP_ALL,
                  'software':('software',
                              'software_version',
                              'software_vendor',
                              'software_url',
                              'software_rrid'),}
    ignore_header = __internal_id_1,
    ignore_alt = 'additional_fields_e_g__minds',
    ignore_match = [['subjects', 'sub-1', {'subject_id': 'sub-1',
                                           'pool_id': 'pool-1',
                                           'reference_atlas': 'Paxinos Rat V3',
                                           'handedness': 'right',}]]
    raw_json_class = rj.RawJsonSubjects
    normalization_class = nml.NormSubjectsFile
    normalize_header = False
    _expect_single = _nsffes

    @property
    def data(self):
        data = super().data
        #breakpoint()
        return data


class SubjectsFilePath(ObjectPath):
    obj = SubjectsFile


SubjectsFilePath._bind_flavours()


class SamplesFile(SubjectsFile):
    """ TODO ... """

    # TODO merge
    __internal_id_1 = object()
    __primary_key = 'primary_key'
    primary_key_rule = __primary_key, ('subject_id', 'sample_id'), '_'.join
    __rename_1 = primary_key_rule[2](('Lab-based schema for identifying each subject',
                                      ('Lab-based schema for identifying each sample, must '
                                       'be unique')))
    renames_header = {__primary_key: 'metadata_element',
                      __rename_1: __internal_id_1,}
    record_type_key_alt = __primary_key
    record_type_key_header = 'metadata_element'
    groups_alt = {'samples': GROUP_ALL,
                  'software':('software',
                              'software_version',
                              'software_vendor',
                              'software_url',
                              'software_rrid'),}
    ignore_header = __internal_id_1,
    ignore_alt = 'additional_fields_e_g__minds',
    ignore_match = [['samples', primary_key_rule[2](('sub-1', 'sub-1_sam-2')),
                     {'subject_id': 'sub-1',
                      'pool_id': 'pool-1',
                      'reference_atlas': 'Paxinos Rat V3',
                      'handedness': 'right',}]]
    normalization_class = nml.NormSamplesFile
    normalize_header = False
    _expect_single = _nsffes


class SamplesFilePath(ObjectPath):
    obj = SamplesFile


SamplesFilePath._bind_flavours()


class ManifestFile(MetadataFile):  # FIXME need a PatternManifestFile I think?
    renames_header = {'filename': 'metadata_element',}
    record_type_key_alt = 'filename'
    record_type_key_header = 'metadata_element'
    groups_alt = {'manifest_records': GROUP_ALL}
    normalize_header = False


class ManifestFilePath(ObjectPath):
    obj = ManifestFile


ManifestFilePath._bind_flavours()


# XXX NOTE *_file is added to all of these keys in DatasetStructure.data
DatasetStructure.sections = {'submission': SubmissionFilePath,
                             'dataset_description': DatasetDescriptionFilePath,
                             'subjects': SubjectsFilePath,
                             'samples': SamplesFilePath,
                             'manifest': ManifestFilePath,}

DatasetStructure._bind_sections()
DatasetStructure._pipeline_stage = DatasetStructure.__name__
DatasetStructure._bind_flavours(pos_helpers=(HasErrors,), win_helpers=(HasErrors,))



from itertools import chain
from augpathlib.exceptions import *


class SparCurError(Exception):
    """ base class for sparcur errors """


class ValidationError(SparCurError):
    def __init__(self, errors):
        self.errors = errors

    def __repr__(self):
        msg = ', '.join([self._format_jsonschema_error(e) for e in self.errors])
        return self.__class__.__name__ + f'({msg})'

    def __str__(self):
        return repr(self)

    def json(self, pipeline_stage_name=None, blame='stage'):
        """ update this to change how errors appear in the validation pipeline """
        skip = 'schema', 'instance', 'context'  # have to skip context because it has unserializable content

        def norm(k, v):
            if k == 'message':
                mess = v
                lm = len(mess)
                dangerzone = 'is not valid under any of the given schemas'
                if (mess.endswith(dangerzone) and
                    lm > 120 and mess.startswith('{')):
                    ld = len(dangerzone)
                    new_mess = (mess[:20] +
                                f' ... {lm - 40 - ld} bytes later ... ' +
                                mess[-(20 + ld):])
                    return new_mess
                else:
                    return mess
            else:
                return v

        return [{k:norm(k, v) if k not in skip else k + ' REMOVED'
                 for k, v in chain(e._contents().items(),
                                   (('pipeline_stage', pipeline_stage_name),
                                    ('blame', blame)))
                 # TODO see if it makes sense to drop these because the parser did know ...
                 if v and k not in skip}
                for e in self.errors]

    @staticmethod
    def _format_jsonschema_error(error):
        """Format a :py:class:`jsonschema.ValidationError` as a string."""
        if error.path:
            dotted_path = ".".join([str(c) for c in error.path])
            return "{path}: {message}".format(path=dotted_path, message=error.message)
        return error.message



class MissingSecretError(SparCurError):
    """ key not in secrets """


class NoFileIdError(SparCurError):
    """ no file_id """


class NoCachedMetadataError(SparCurError):
    """ there is no cached metadata """


class AlreadyInProjectError(SparCurError):
    """fatal: already in a spc project {}"""
    def __init__(self, message=None):
        if message is None:
            more = '(or any of the parent directories)' # TODO filesystem boundaries ?
            self.message = self.__doc__.format(more)


class NotInDatasetError(SparCurError):
    """ trying to run a comman on a dataset when not inside one """


class NotBootstrappingError(SparCurError):
    """ Trying to run bootstrapping only code outside of a bootstrap """


class EncodingError(SparCurError):
    """ Some encoding error has occured in a file """


class FileTypeError(SparCurError):
    """ File type is not allowed """


class NoDataError(SparCurError):
    """ There was no data in the file (not verified with stat)
        FIXME HACK workaround for bad handling of empty sheets in byCol """


class BadDataError(SparCurError):
    """ something went wrong """


class MalformedHeaderError(BadDataError):
    """ Bad header """


class CouldNotNormalizeError(SparCurError):
    """ signal that a value could not be normalized """


class TabularCellError(SparCurError):
    """ signal that an error has occured in a particular cell """
    def __init__(self, msg, *, value=None, location=None):
        self.value = value
        self.location = location  # usually set after the fact
        if location:
            msg = f'{msg} at {location}'

        super().__init__(msg)


class LengthMismatchError(SparCurError):
    """ lenghts of iterators for a zipeq do not match """


class LengthMismatchError(SparCurError):
    """ lenghts of iterators for a zipeq do not match """


class NotApplicableError(SparCurError):
    """ There are a number of cases where N/A values should
        be treated as errors that need to be caugt so that
        the values can be cut out entirely. """


class SubPipelineError(SparCurError):
    """ There was an error in a subpipeline. """


class NoTripleError(SparCurError):
    """ an evil hack to prevent export of a triple THANKS STUPID DOI DECISIONS """


class LostChildError(SparCurError):
    """ someone attempting to upload a child to the wrong parent """


class NetworkFailedForPathError(SparCurError):
    """ the network failed while trying to retrieve a specfic path """



import unittest
from .common import template_root, project_path
from sparcur.curation import Summary
from sparcur.blackfynn_api import FakeBFLocal


class TestSummary(unittest.TestCase):
    def setUp(self):
        project_path.cache.anchorClassHere(remote_init=False)
        project_path._remote_class._api = FakeBFLocal(project_path.cache.id, project_path.cache)
        self.s = Summary(project_path)
        self.s._n_jobs = 1
        self.s.setup(local_only=True)

    def test_data(self):
        self.s.data()



import io
import os
import json
import types
import asyncio
from copy import deepcopy
#from nibabel import nifti1
#from pydicom import dcmread
#from scipy.io import loadmat
import yaml
import boto3
import botocore
import requests
from requests import Session
from requests.exceptions import HTTPError, ConnectionError
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from joblib import Parallel, delayed
if 'BLACKFYNN_LOG_LEVEL' not in os.environ:
    # silence agent import warning
    os.environ['BLACKFYNN_LOG_LEVEL'] = 'CRITICAL'
from blackfynn import Blackfynn, Collection, DataPackage, Organization, File
from blackfynn import Dataset, BaseNode
from blackfynn.models import BaseCollection
from blackfynn import base as bfb
from blackfynn.api import transfers
from blackfynn.api.data import PackagesAPI
from pyontutils.utils import Async, deferred, async_getter, chunk_list
from pyontutils.iterio import IterIO
from sparcur import exceptions as exc
from sparcur.core import log, lj
from sparcur.metastore import MetaStore
from .config import auth


def upload_fileobj(
        file,  # aka Path
        s3_host,
        s3_port,
        s3_bucket,
        s3_keybase,
        region,
        access_key_id,
        secret_access_key,
        session_token,
        encryption_key_id,
        upload_session_id=None,
        ):
    """ streaming upload
        the object passed in as 'file'
        doesn't have to be Path at all
        it just needs to implement the following methods
        `name`, `size`, and `data`
    """
    local_path = file

    try:
        # account for dev connections
        resource_args = {}
        config_args = dict(signature_version='s3v4')
        if 'amazon' not in s3_host.lower() and len(s3_host)!=0:
            resource_args = dict(endpoint_url="http://{}:{}".format(s3_host, s3_port))
            config_args = dict(s3=dict(addressing_style='path'))

        # connect to s3
        session = boto3.session.Session()
        s3 = session.client('s3',
            region_name = region,
            aws_access_key_id = access_key_id,
            aws_secret_access_key = secret_access_key,
            aws_session_token = session_token,
            config = botocore.client.Config(**config_args),
            **resource_args
        )

        # s3 key
        s3_key = '{}/{}'.format(s3_keybase, local_path.name)

        # override seek to raise an IOError so
        # we don't get a TypeError
        # FIXME IterIO stores a buffer of the whole generator >_<
        f = IterIO(local_path.data, sentinel=b'')
        def _seek(self, *args):
            raise IOError('nope')
        f.seek = _seek

        # upload file to s3
        s3.upload_fileobj(
            Fileobj=f,  # FIXME checksumming wrapper probably ...
            Bucket=s3_bucket,
            Key=s3_key,
            #Callback=progress,
            ExtraArgs=dict(
                ServerSideEncryption="aws:kms",
                SSEKMSKeyId=encryption_key_id,
                #Metadata=checksums,  # hca does it this way
                # annoyingly this means that you have to read the file twice :/
            ))

        return s3_key

    except Exception as e:
        log.error(e)
        raise e


transfers.upload_file = upload_fileobj


def check_files(files):
    """ of course they don't exist """


transfers.check_files = check_files


#_orgid = auth.get('blackfynn-organization')  # HACK
# this works, but then we have to do the multipart upload dance
# which the bf client does not implement at the moment
def get_preview(self, files, append):
    paths = files

    params = dict(
        append = append,
        # dataset_id=?,
    )

    payload = { "files": [
        {
            "fileName": p.name,
            "size": p.size,
            "uploadId": i,
            "processing": True,
        } for i, p in enumerate(paths)
    ]}

    # current web dance
    # /upload/preview/organizations/{_orgid}?append= append ?dataset_id= integer???
    # /upload/fineuploaderchunk/organizations/{_orgid}/id/{importId}?multipartId= from-prev-resp
    # /upload/complete/organizations/{_orgid}/id/{importId}?datasetId=
    response = self._post(
        endpoint=self._uri('/files/upload/preview'),
        #endpoint=self._uri(f'/upload/preview/organizations/{_orgid}'),
        params=params,
        json=payload,
        )

    import_id_map = dict()
    for p in response.get("packages", list()):
        import_id = p.get("importId")
        warnings = p.get("warnings", list())
        for warning in warnings:
            logger.warn("API warning: {}".format(warning))
        for f in p.get("files", list()):
            index = f.get("uploadId")
            import_id_map[paths[index]] = import_id
    return import_id_map


transfers.IOAPI.get_preview = get_preview


def init_file(self, filename):
    """ not using any of this """


transfers.UploadManager.init_file = init_file


# monkey patches


@property
def patch_session(self):
    """
    Make requests-futures work within threaded/distributed environment.
    """
    if self._session is None:
        self._session = Session()
        self._set_auth(self._token)

        # Enable retries via urllib
        adapter = HTTPAdapter(
            pool_connections=1000,  # wheeee
            pool_maxsize=1000,  # wheeee
            max_retries=Retry(
                total=self.settings.max_request_timeout_retries,
                backoff_factor=.5,
                status_forcelist=[502, 503, 504] # Retriable errors (but not POSTs)
            )
        )
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)

    return self._session

# monkey patch to avoid session overflow during async
bfb.ClientSession.session = patch_session


def get(self, id, update=True):
    return self._api.core.get(id, update=update)


# monkey patch Blackfynn so that it doesn't eat errors and hide their type
Blackfynn.get = get


def download(self, destination):
    """ remove prefix functionality since there are filenames without extensions ... """
    if self.type=="DirectoryViewerData":
        raise NotImplementedError("Downloading S3 directories is currently not supported")

    if os.path.isdir(destination):
        # destination dir
        f_local = os.path.join(destination, os.path.basename(self.s3_key))
    else:
        # exact location
        f_local = destination

    r = requests.get(self.url, stream=True)
    with io.open(f_local, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk: f.write(chunk)

    # set local path
    self.local_path = f_local

    return f_local

# monkey patch File.download to
File.download = download


# package meta
def id_to_type(id):
    if id.startswith('N:package:'):
        return DataPackage
    elif id.startswith('N:collection:'):
        return Collection
    elif id.startswith('N:dataset:'):
        return Dataset
    elif id.startswith('N:organization:'):
        return Organization


class FakeBFile(File):
    """ Fake file to simplify working with package metadata """

    id = None  # unforunately these don't seem to have made it through

    def __init__(self, package, **kwargs):
        self.package = package
        self._json = kwargs
        def move(*tuples):
            for f, t in tuples:
                kwargs[t] = kwargs.pop(f)

        move(('createdAt', 'created_at'),
             ('updatedAt', 'updated_at'),
             ('fileType', 'type'),
             ('packageId', 'pkg_id'),
        )

        if 'size' not in kwargs:
            kwargs['size'] = None  # if we have None on a package we know it is not zero

        if 'checksum' in kwargs:
            cs = kwargs['checksum']
            kwargs['chunksize'] = cs['chunkSize']
            kwargs['checksum'] = cs['checksum']  # overwrites but ok
        else:
            kwargs['checksum'] = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def state(self):
        return self.package.state

    @property
    def owner_id(self):
        return self.package.owner_id

    @property
    def dataset(self):
        return self.package.dataset

    @property
    def parent(self):
        # cut out the middle man (hopefully)
        return self.package.parent

    def __repr__(self):
        return ('<' +
                self.__class__.__name__ +
                ' '.join(f'{k}={v}' for k, v in self._json.items()) +
                '>')


    
# monkey patch for PackagesAPI.get
from future.utils import string_types
def get(self, pkg, include='files,source'):
    """
    Get package object

    pkg:     can be DataPackage ID or DataPackage object.
    include: list of fields to force-include in response (if available)
    """
    pkg_id = self._get_id(pkg)

    params = {'includeAncestors': 'true'}
    if include is not None:
        if isinstance(include, string_types):
            params.update({'include': include})
        elif hasattr(include, '__iter__'):
            params.update({'include': ','.join(include)})

    resp = self._get(self._uri('/{id}', id=pkg_id), params=params)

    # TODO: cast to specific DataPackages based on `type`
    pkg = self._get_package_from_data(resp)
    pkg._resp = resp
    #log.debug(lj(resp.json()))
    return pkg

PackagesAPI.get = get


@property
def packages(self, pageSize=1000, includeSourceFiles=True):
    yield from self._packages(pageSize=pageSize, includeSourceFiles=includeSourceFiles)


def packagesByName(self, pageSize=1000, includeSourceFiles=True, filenames=tuple()):
    if filenames:
        for filename in filenames:
            yield from self._packages(pageSize=pageSize, includeSourceFiles=includeSourceFiles, filename=filename)
    else:
        yield from self._packages(pageSize=pageSize, includeSourceFiles=includeSourceFiles)


def _packages(self, pageSize=1000, includeSourceFiles=True, raw=False, latest_only=False, filename=None):
    """ python implementation to make use of /dataset/{id}/packages """
    remapids = {}
    def restructure(j):
        """ restructure package json to match what api needs? """
        # FIXME something's still wonky here
        c = j['content']
        c['int_id'] = c['id']
        c['id'] = c['nodeId']  # FIXME indeed packages do seem to be missing ids!?
        remapids[c['int_id']] = c['id']
        c['int_datasetId'] = c['datasetId']
        c['datasetId'] = c['datasetNodeId']
        if 'parentId' in c:
            pid = c['parentId']
            c['parent'] = remapids[pid]  # key error to signal out of order
            #if pid in remapids:
            #else:
                #c['parent'] = f'WTF ERROR: {pid}'
                #print('wtf', pid, c['id'], c['datasetId'])
            #else:
                #c['parent'] = remapids['latest']
        return j

    index = {self.id:self}  # make sure that dataset is in the index
    session = self._api.session
    #cursor
    #pageSize
    #includeSourceFiles
    #types
    #filename
    filename_args = f'&filename={filename}' if filename is not None else ''
    cursor_args = ''
    out_of_order = []
    while True:
        try:
            resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/packages?'
                            f'pageSize={pageSize}&'
                            f'includeSourceFiles={str(includeSourceFiles).lower()}'
                            f'{filename_args}'
                            f'{cursor_args}')
        except requests.exceptions.RetryError as e:
            log.exception(e)
            # sporadic 504 errors that we probably need to sleep on
            breakpoint()
            raise e

        #print(resp.url)
        if resp.ok:
            j = resp.json()
            packages = j['packages']
            if raw:
                yield from packages
                if latest_only:
                    break
                else:
                    continue

            if out_of_order:
                packages += out_of_order
                # if a parent is on the other side of a
                # pagination boundary put the children
                # at the end and move on
            out_of_order = [None]
            while out_of_order:
                #log.debug(f'{out_of_order}')
                if out_of_order[0] is None:
                    out_of_order.remove(None)
                elif packages == out_of_order:
                    if filename is not None:
                        out_of_order = None
                    elif 'cursor' not in j:
                        raise RuntimeError('We are going nowhere!')
                    else:
                        # the missing parent is in another castle!
                        break
                else:
                    packages = out_of_order
                    out_of_order = []
                for count, package in enumerate(packages):
                    if isinstance(package, dict):
                        id = package['content']['nodeId']
                        name = package['content']['name']
                        bftype = id_to_type(id)
                        dcp = deepcopy(package)
                        try:
                            #if id.startswith('N:package:'):
                                #log.debug(lj(package))
                            rdp = restructure(dcp)
                        except KeyError as e:
                            if out_of_order is None:  # filename case
                                # parents will simply not be listed
                                # if you are using filename then beware
                                if 'parentId' in dcp['content']:
                                    dcp['content'].pop('parentId')
                                rdp = restructure(dcp)
                            else:
                                out_of_order.append(package)
                                continue

                        bfobject = bftype.from_dict(rdp, api=self._api)
                        if name != bfobject.name:
                            log.critical(f'{name} != {bfobject.name}')
                        bfobject._json = package
                        bfobject.dataset = index[bfobject.dataset]
                    else:
                        bfobject = package

                    if isinstance(bfobject.parent, str) and bfobject.parent in index:
                        parent = index[bfobject.parent]
                        if parent._items is None:
                            parent._items = []
                        parent.items.append(bfobject)
                        bfobject.parent = parent
                        # only put objects in the index when they have a parent
                        # that is a bfobject, this ensures that you can always
                        # recurse to base once you get an object from this function
                        index[bfobject.id] = bfobject
                        if parent.state == 'DELETING':
                            if not bfobject.state == 'DELETING':
                                bfobject.state = 'PARENT-DELETING'
                        elif parent.state == 'PARENT-DELETING':
                            if not bfobject.state == 'DELETING':
                                bfobject.state = 'PARENT-DELETING'

                        yield bfobject  # only yield if we can get a parent
                    elif out_of_order is None:  # filename case
                        yield bfobject
                    elif bfobject.parent is None:
                        # both collections and packages can be at the top level
                        # dataset was set to its bfobject repr above so safe to yield
                        if bfobject.dataset is None:
                            log.debug('No parent no dataset\n'
                                      + json.dumps(bfobject._json, indent=2))
                        index[bfobject.id] = bfobject
                        yield bfobject
                    else:
                        out_of_order.append(bfobject)
                        continue

                    if isinstance(bfobject, DataPackage):
                        bfobject.fake_files = []
                        if 'objects' not in bfobject._json:
                            log.error(f'{bfobject} has no files!??!')
                        else:
                            for i, source in enumerate(bfobject._json['objects']['source']):
                                # TODO package id?
                                if len(source) > 1:
                                    log.info(f'more than one key in source {sorted(source)}')

                                ff = FakeBFile(bfobject, **source['content'])
                                bfobject.fake_files.append(ff)
                                yield ff

                                if i == 1:  # only log once
                                    log.critical(f'MORE THAN ONE FILE IN PACKAGE {bfobject.id}')

            if 'cursor' in j:
                cursor = j['cursor']
                cursor_args = f'&cursor={cursor}'
            else:
                break

        else:
            break


@property
def packageTypeCounts(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/packageTypeCounts')
    if resp.ok:
        j = resp.json()
        return j
    else:
        resp.raise_for_status()


# monkey patch Dataset to implement packages endpoint
Dataset._packages = _packages
Dataset.packages = packages
Dataset.packagesByName = packagesByName
Dataset.packageTypeCounts = packageTypeCounts


@property
def users(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/collaborators/users')
    return resp.json()


# monkey patch Dataset to implement teams endpoint
Dataset.users = users


@property
def teams(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/collaborators/teams')
    return resp.json()


# monkey patch Dataset to implement teams endpoint
Dataset.teams = teams


@property
def doi(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/doi')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code != 404:
            log.warning(f'{self} doi {resp.status_code}')


# monkey patch Dataset to implement doi endpoint
Dataset.doi = doi


def delete(self):
    """ actually delete """
    session = self._api.session
    resp = session.delete(f'https://api.blackfynn.io/datasets/{self.id}')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code != 404:
            log.warning(f'{self} issue deleting {resp.status_code}')


# monkey patch Dataset to implement delete
Dataset.delete = delete


@property
def contributors(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/contributors')
    return resp.json()


# monkey patch Dataset to implement contributors endpoint
Dataset.contributors = contributors


@property
def banner(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/banner')
    return resp.json()


# monkey patch Dataset to implement banner endpoint
Dataset.banner = banner


@property
def readme(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/readme')
    return resp.json()


# monkey patch Dataset to implement readme endpoint
Dataset.readme = readme


@property
def status_log(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}/status-log')
    return resp.json()


# monkey patch Dataset to implement status_log endpoint
Dataset.status_log = status_log


@property
def meta(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/datasets/{self.id}')
    return resp.json()


# monkey patch Dataset to implement just the dataset metadata endpoint
Dataset.meta = meta


@property
def teams(self):
    session = self._api.session
    resp = session.get(f'https://api.blackfynn.io/organizations/{self.id}/teams')
    return resp.json()


# monkey patch Organization to implement teams endpoint
Organization.teams = teams


prefix = 'https://app.blackfynn.io/'
lp = len(prefix)
def destructure_uri(uri):
    if not uri.startswith(prefix):
        return None  # TODO less cryptic return value or exception

    suffix = uri[lp:]
    for maybe_id in suffix.split('/'):
        if maybe_id.startswith('N:'):
            yield maybe_id

    'N:organization:89dfbdad-a451-4941-ad97-4b8479ed3de4'
    '/datasets/'
    'N:dataset:bedda0db-c275-4d79-87ce-fc7bf1e11600'
    '/files'


def make_filename(file):
    # we have to do this because the type on the package is unreliable
    # so we have to get it from s3 because that is what will roundtrip
    _, file_name_s3 = file.s3_key.rsplit('/', 1)
    return file_name_s3

def process_package(package, path, doasync=True):
    # packages are the souless shells of files
    # you can retrieve them quickly, but they have no substance

    if not path.exists():
        path.mkdir(parents=True)

    if isinstance(package, Collection):
        npath = path / package.name
        if doasync:  # FIXME recursive async doesn't quite work
            for lst in Async(debug=True)(deferred(pp)(npackage, npath)
                                        for npackage in package):
                yield from lst
        else:
            yield from (process_package(npackage, npath, False)
                        for npackage in package)

    else:
        for file in package.files:
            file_name = make_filename(file)
            log.debug(file_name)
            file_path = path / file_name
            yield file, file_path


def pp(package, path, doasync=True):
    return list(process_package(package, path, doasync))


def inner(thing):
    """ one level """
    if isinstance(thing, DataPackage):
        return thing,
    else:
        return list(thing)

def outer(dataset):
    return [e for t in Async()(deferred(inner)(thing) for thing in dataset) for e in t]

def heh(c):
    #print(c)
    return list(c)

def asynchelper(chunk):
    import asyncio
    asyncio.set_event_loop(asyncio.new_event_loop())
    wat = async_getter(heh, [(c,) for c in chunk]) 
    log.debug('chunkdone')
    return [e for t in wat for e in t]

def get_packages(package_or_collection, path):
    """ flatten collections into packages """
    if isinstance(package_or_collection, Collection):
        npath = path / NormFolder(package_or_collection.name)
        yield package_or_collection, path
        for npc in package_or_collection:
            yield from get_packages(npc, npath)
    else:
        log.debug(f'{path} {package_or_collection}')
        yield package_or_collection, path


def get_packages_(dataset):
    """ flatten collections into packages """

    hrm1 = outer(dataset)
    print(len(dataset))
    print(len(hrm1))
    hrm2 = outer(hrm1)
    print(len(hrm2))
    chunks = chunk_list([c for c in hrm2 if isinstance(c, Collection)], 1000)
    #wat = async_getter(heh, [(c,) for c in hrm2 if isinstance(c, Collection)][:1000]) 
    Parallel(n_jobs=8, backend="threading")(delayed(asynchelper)(chunk) for chunk in chunks)
    breakpoint()
    return
    if collector is None:
        collector = []

    files = []
    folders = []
    if isinstance(package_or_collection, Collection):
        list(package_or_collection)
    else:
        files.append(package_or_collection, path)

def blah():
    if isinstance(package_or_collection, Collection):
        [t for npc in package_or_collection
         for t in (list(npc) if isinstance(npc, Collection) else (npc,))]
    else:
        yield package_or_collection, path

def pkgs_breadth(package_or_collection, path):
    print(path)
    if isinstance(package_or_collection, Collection):
        npath = path / NormFolder(package_or_collection.name)
        to_iter = list(package_or_collection)
        #print(to_iter)
        coll = []
        for thing in Async()(deferred(get_packages)(npc, npath) for npc in to_iter):
            coll.append(thing)
            #yield from thing

        yield coll
        Async()(deferred(list)(t) for t in thing)
        #yield (list(c) for c in coll)
        #[v for h in hrm for v in h]

        #for npc in package_or_collection:
            #for poc_p in 
            #yield from get_packages(npc, npath)
    else:
        yield package_or_collection, path

def pkgs_depth(mess):
    hrm = Async()(deferred(list)(t) for t in mess)

class FakeFile:
    def __init__(self, package):
        self.s3_key = '/' + package.name
        self.size = -1
        self.id = -1
        self.pkg_id = package.id
        self.created_at = package.created_at
        self.updated_at = package.updated_at

def gfiles(package, path):
    # print(p.name)
    # the fanout is horrible ...
    for i in range(100):
        try:
            return path, package.files
        except HTTPError as e:
            status_code = e.response.status_code
            print(e)
            asyncio.sleep(2)
    else:
        return path, FakeFile(package)

def unlink_fakes(attrs, fake_paths, metastore):
    for fpath in fake_paths:
        fattrs = {k:v for k, v in norm_xattrs(fpath.xattrs()).items() if k != 'bf.error'}
        if fattrs == attrs:
            fpath.unlink()
            metastore.remove(fpath)
        else:
            print('WARNING: fake xattrs and real xattrs do not match!', attrs, fattrs)


def norm_xattrs(attrs):
    # FIXME update Path.xattrs to accept a normalization function
    out = {}
    ints = 'bf.file_id', 'bf.size'
    strings = 'bf.id', 'bf.created_at', 'bf.updated_at'
    for k, v in attrs.items():
        if isinstance(k, bytes):
            k = k.decode()  # still handled by Path

        if k in strings:
            v = v.decode('utf-8')
        elif k in ints:
            v = int(v)

        out[k] = v

    return out


def make_file_xattrs(file):
    return {
        'bf.id':file.pkg_id,
        'bf.file_id':file.id,
        'bf.size':file.size,
        'bf.created_at':file.updated_at,
        'bf.updated_at':file.created_at,
        'bf.checksum':'',  # TODO
        # 'bf.old_id': '',  # TODO does this work? also naming previous_id, old_version_id etc...
    }


def make_folder_xattrs(folder):
    """ I weep for the world where files and folders were the same thing.
        I just want this file though! NO YOU MUST TAKE IT ALL.
        Does have some usability issues, but would have made metadata
        SO much easier to deal with and explain to people. Of course
        there are some major antipatterns for a system that can do that. """
    out = {'bf.id':folder.id}
    if not folder.id.startswith('N:organization:'):
        out.update({'bf.created_at':folder.updated_at,
                    'bf.updated_at':folder.created_at,})

    return out


def fetch_file(file_path, file, metastore, limit=False, overwrite=False):
    if not file_path.parent.exists():
        file_path.parent.mkdir(parents=True, exist_ok=True)

    fake_paths = list(file_path.parent.glob(file_path.name + '.fake*'))

    if file_path.exists() and not overwrite:
        print('already have', file_path)
        attrs = norm_xattrs(file_path.xattrs())
        unlink_fakes(attrs, fake_paths, metastore)
        return

    limit_mb = 2
    file_mb = file.size / 1024 ** 2
    skip = 'jpeg', 'jpg', 'tif', 'png'
    file_xattrs = make_file_xattrs(file)

    if (not limit or
        (file_mb < limit_mb and
         (not file_path.suffixes or
          file_path.suffixes[0][1:].lower() not in skip))):
        print('fetching', file)
        for i in range(4):  # try 4 times
            try:
                # FIXME I think README_README is an s3_key related error
                file.download(file_path.as_posix())
                file_path.setxattrs(file_xattrs)
                #metastore.setxattrs(file_path, file_xattrs)  # FIXME concurrent access kills this
                attrs = norm_xattrs(file_path.xattrs())  # yes slow, but is a sanity check
                # TODO validate the checksum when we get it
                unlink_fakes(attrs, fake_paths, metastore)
                return
            except (HTTPError, ConnectionError) as e:
                error = str(e)
                status_code = e.response.status_code
                asyncio.sleep(3)
        else:
            print(error)
            error_path = file_path.with_suffix(file_path.suffix + '.fake.ERROR')
            error_path.touch()
            file_xattrs['bf.error'] = str(status_code)
            error_path.setxattrs(file_xattrs)
            #metastore.setxattrs(error_path, file_xattrs)
    else:
        fsize = str(int(file_mb)) + 'M' if file_mb >= 1 else str(file.size // 1024) + 'K'
        fakepath = file_path.with_suffix(file_path.suffix + '.fake.' + fsize)
        fakepath.touch()
        fakepath.setxattrs(file_xattrs)
        #metastore.setxattrs(fakepath, file_xattrs)


def make_files_meta(collection):
    # TODO file fetching status? file hash?
    return {make_filename(file):[package.id, file.id, file.size]
            for package in collection
            if isinstance(package, DataPackage)
            for file in package.files}


class NormFolder(str):
    def __new__(cls, value):
        return str.__new__(cls, cls.normalize(value))

    @classmethod
    def normalize(cls, value):
        # I CAN'T BELIEVE YOU'VE DONE THIS
        return value.replace('/', '_')


def make_folder_and_meta(parent_path, collection, metastore):
    """ maps to collection dataset or organization """
    folder_path = parent_path / NormFolder(collection.name)
    #meta_file = folder_path / collection.id
    #if isinstance(collection, Organization):
        #files_meta = {}
    #else:
        #files_meta = make_files_meta(collection)
    folder_path.mkdir(parents=True, exist_ok=True)
    attrs = make_folder_xattrs(collection)
    folder_path.setxattrs(attrs)
    # sadly xattrs are easy to accidentally zap :/
    #metastore.setxattr(folder_path, 'bf.id', collection.id)
    #with open(meta_file, 'wt') as f:
        #yaml.dump(files_meta, f, default_flow_style=False)


def get_file_by_id(get, file_path, pid, fid):
    package = get(pid)
    if package is None:
        print('WARNING package does not exist', file_path, pid, fid)
        return None, None

    for f in package.files:
        if f.id == fid:
            return file_path, f
    else:
        print('WARNING file does not exist', file_path, pid, fid)
        return None, None


class PackageMeta:

    class PackageMetaError(Exception):
        """ Ya done goofed part 3 """

    class NotDataPackage(Exception):
        """ only packages have """

    class NoPackageMetaError(Exception):
        """ no package meta """

    def __init__(self, hbfo):
        bfobject = hbfo.bfobject
        if not type(bfobject) == DataPackage:
            msg = f'{bfobject} is not a DataPackage'
            raise self.NotDataPackage(msg)

        elif not hasattr(bfobject, '_json'):
            msg = f'{bfobject} was not spawned from /datasets/{{id}}/packages'
            raise self.NoPackageMetaError(msg)

        self.hbfo = hbfo
        self.bfobject = bfobject
        self.json = hbfo.bfobject._json

    @property
    def size(self):
        self.json['objects']['source']


class BFLocal:

    class NoBfMeta(Exception):
        """ There is not bf id for this file. """

    def __init__(self, project_id, anchor=None):
        # no changing local storage prefix in the middle of things
        # if you want to do that create a new class

        self._project_id = project_id
        self.bf = self._get_connection(self._project_id)

        self.organization = self.bf.context
        self.project_name = self.bf.context.name

    def _get_connection(self, project_id):
        try:
            return Blackfynn(api_token=auth.user_config.secrets('blackfynn', self._project_id, 'key'),
                             api_secret=auth.user_config.secrets('blackfynn', self._project_id, 'secret'))
        except KeyError as e:
            msg = f'need record in secrets for blackfynn organization {self._project_id}'
            raise exc.MissingSecretError(msg) from e

    def __getstate__(self):
        state = self.__dict__
        state.pop('bf')  # does not picle well due to connection
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.bf = self._get_connection(self._project_id)

    @property
    def root(self):
        return self.organization.id

    def create_package(self, local_path):
        pkg = DataPackage(local_path.name, package_type='Generic')  # TODO mimetype -> package_type ...
        pcache = local_path.parent.cache
        pkg.dataset = pcache.dataset.id
        if pcache.id != pkg.dataset:
            pkg.parent = pcache.id

        # FIXME this seems to create an empty package with no files?
        # have to do aws upload first or something?
        return self.bf._api.packages.create(pkg)

    def get(self, id, attempt=1, retry_limit=3):
        log.debug('We have gone to the network!')
        if id.startswith('N:dataset:'):
            try:
                thing = self.bf.get_dataset(id)  # heterogenity is fun!
            except Exception as e:
                if 'No dataset matching name or ID' in str(e):
                    # sigh no error types
                    raise exc.NoRemoteFileWithThatIdError(id) from e
                else:
                    raise e

        elif id.startswith('N:organization:'):
            if id == self.organization.id:
                return self.organization  # FIXME staleness?
            else:
                # if we start form local storage prefix for everything then
                # this would work
                raise BaseException('TODO org does not match need other api keys.')
        else:
            try:
                thing = self.bf.get(id)
            except requests.exceptions.HTTPError as e:
                resp = e.response
                if resp.status_code == 404:
                    msg = f'{resp.status_code} {resp.reason!r} when fetching {resp.url}'
                    raise exc.NoRemoteFileWithThatIdError(msg) from e

                log.exception(e)
                thing = None
            except bfb.UnauthorizedException as e:
                log.error(f'Unauthorized to access {id}')
                thing = None

        if thing is None:
            if attempt > retry_limit:
                raise exc.NoMetadataRetrievedError(f'No blackfynn object retrieved for {id}')
            else:
                thing = self.get(id, attempt + 1)

        return thing

    def get_file_url(self, id, file_id):
        resp = self.bf._api.session.get(f'https://api.blackfynn.io/packages/{id}/files/{file_id}')
        if resp.ok:
            resp_json = resp.json()
        elif resp.status_code == 404:
            msg = f'{resp.status_code} {resp.reason!r} when fetching {resp.url}'
            raise exc.NoRemoteFileWithThatIdError(msg)
        else:
            resp.raise_for_status()

        try:
            return resp_json['url']
        except KeyError as e:
            log.debug(lj(resp_json))
            raise e

    def recover_meta(self, path):
        pattrs = norm_xattrs(path.parent.xattrs())
        codid = pattrs['bf.id']
        thing = self.get(codid)
        if thing is None:
            return {}
            raise BaseException('What are you doing!??!?!')

        test_path = path
        while '.fake' in test_path.suffixes:
            test_path = test_path.with_suffix('')

        for poc in thing:
            if NormFolder(poc.name) == test_path.stem:  # FIXME
                if isinstance(poc, Collection):
                    return {'bf.id':poc.id}

                for file in poc.files:  # FIXME files vs folders
                    filename = make_filename(file)
                    if filename == test_path.name:
                        return make_file_xattrs(file)

        else:
            raise BaseException(f'\'{path}\' could not recover meta!')

    def get_file_meta(self, path):
        attrs = norm_xattrs(path.xattrs())
        if 'bf.id' not in attrs:
            # TODO maintain a single backup mapping of xattrs to paths
            # and just use the xattrs for performance
            attrs = self.metastore.xattrs(path)
            if attrs:
                # TODO checksum ... (sigh git)
                path.setxattrs(attrx)
                attrs = norm_xattrs(path.xattrs())
            else:
                raise self.NoBfMeta

        pid = attrs['bf.id']
        if pid.startswith('N:package:'):
            fid = int(attrs['bf.file_id'])
            file_path = path
            while '.fake' in file_path.suffixes:
                file_path = file_path.with_suffix('')

            return file_path, pid, fid
        else:
            log.warning(f'what is going on with {path} {attrs}')

    def fetch_path(self, path, overwrite=False):
        """ Fetch individual big files.
            `path` argument must be to a fake file which has the meta stored in xattrs """
        # automatic async function application inside a list comp ... would be fun
        fetch_file(*get_file_by_id(self.bf.get, *self.get_file_meta(path)), self.metastore, overwrite=overwrite)

    def fetch_errors(self):
        bfiles = {fp:f for fp, f in
                  Async()(deferred(get_file_by_id)(self.bf.get, file_path, pid, fid)
                          for file_path, pid, fid in self.error_meta)}
        Async()(deferred(fetch_file)(filepath, file, self.metastore) for filepath, file in bfiles.items() if not filepath.exists())

    def file_fetch_dict(self, packages):
        return {folder_path / make_filename(file):file
                for folder_path, files in
                Async()(deferred(gfiles)(package, path) for package, path in packages)
                for file in files}


class OldStuff:
    def populate_metastore(self):
        """ This should be run after find_missing_meta. """
        # FIXME need a way to delete files
        all_attrs = {path:norm_xattrs(path.xattrs()) for path in self.project_path.rglob('*')
                     if '.git' not in path.as_posix()}
        bad = [path for path, attrs in all_attrs.items() if not attrs]
        if bad:
            log.warning(f'{bad} is missing meta, run find_missing_meta')
            all_attrs = {p:a for p, a in all_attrs.items() if a}

        self.metastore.bulk(all_attrs)

    def find_missing_meta(self):
        for path in self.project_path.rglob('*'):
            if '.git' in path.as_posix():
                continue
            attrs = norm_xattrs(path.xattrs())
            if not attrs:
                log.warning(f'Found path with missing metadata {path}')
                attrs = self.metastore.xattrs(path)
                if not attrs:
                    log.error('No local metadata was found for {path}')
                    attrs = self.recover_meta(path)

                path.setxattrs(attrs)
                # TODO checksum may no longer match since we changed it


class FakeBFLocal(BFLocal):
    class bf:
        """ tricksy hobbitses """
        @classmethod
        def get_dataset(cls, id):
            #import inspect
            #stack = inspect.stack(0)
            #breakpoint()
            #return CacheAsDataset(id)
            class derp:
                """ yep, this is getting called way down inside added
                    in the extras pipeline :/ """
                doi = None
                status = 'FAKE STATUS ;_;'
            return derp

    def __init__(self, project_id, anchor):
        self.organization = CacheAsBFObject(anchor)  # heh
        self.project_name = anchor.name
        self.project_path = anchor.local


class CacheAsBFObject(BaseNode):
    def __init__(self, cache):
        self.cache = cache
        self.id = cache.id
        self.cache.meta

    @property
    def name(self):
        return self.cache.name

    @property
    def parent(self):
        parent_cache = self.cache.parent #.local.parent.cache
        if parent_cache is not None:
            return self.__class__(parent_cache)

    @property
    def parents(self):
        parent = self.parent
        while parent:
            yield parent
            parent = parent.parent

    def __iter__(self):
        for c in self.cache.local.children:
            cache = c.cache
            if cache.is_dataset():
                child = CacheAsDataset(cache)
            elif cache.is_organization():
                child = CacheAsOrganization(cache)
            elif cache.is_dir():
                child = CacheAsCollection(cache)
            elif cache.is_broken_symlink():
                child = CacheAsFile(cache)

            yield child

    @property
    def members(self):
        return []


class CacheAsFile(CacheAsBFObject, File): pass
class CacheAsCollection(CacheAsBFObject, Collection): pass
class CacheAsDataset(CacheAsBFObject, Dataset):
    """ yep, this is getting called way down inside added
        in the extras pipeline :/ """
    doi = None
    status = 'FAKE STATUS ;_;'
    @property
    def created_at(self): return self.cache.meta.created
    @property
    def updated_at(self): return self.cache.meta.updated
    @property
    def owner_id(self): return self.cache.meta.user_id
class CacheAsOrganization(CacheAsBFObject, Organization): pass


def mvp():
    """ In order to be performant for large numbers of packages we have
        to get all the packages first and then async retrieve all the files
    """
    from sparcur import paths
    local_storage_prefix = paths.Path('~/files/blackfynn_local/').expanduser()
    bf = Blackfynn(api_token=auth.dynamic_config.secrets('blackfynn-mvp-key'),
                    api_secret=auth.dynamic_config.secrets('blackfynn-mvp-secret'))


    ds = bf.datasets()
    useful = {d.id:d for d in ds}  # don't worry, I've made this mistake too

    project_name = bf.context.name

    helm = useful['N:dataset:bedda0db-c275-4d79-87ce-fc7bf1e11600']
    helmr = useful['N:dataset:d412a972-870c-4b63-9865-8d790065bd43']
    datasets = helm, helmr  # TODO add more datasets here
    packages = []
    for dataset in datasets:
        dataset_name = dataset.name
        ds_path = local_storage_prefix / project_name / dataset_name
        for package_or_collection in dataset:
            packages.extend(get_packages(package_or_collection, ds_path))

    bfiles = {folder_path / make_filename(file):file
              for folder_path, files in
              Async()(deferred(gfiles)(package, path) for package, path in packages)
              if files is not None  # FIXME how to deal with this!?
              for file in files}

    # beware that this will send as many requests as it can as fast as it can
    # which is not the friendliest thing to do to an api
    Async()(deferred(fetch_file)(*fpf, self.metastore) for fpf in bfiles.items() if not fp.exists())
    self.populate_metastore()  # FIXME workaround for concurrent access issues, probably faster, but :/

    return bf, bfiles


def process_files(bf, files):
    from IPython import embed
    niftis = [nifti1.load(f.as_posix()) for f in files if '.nii' in f.suffixes]
    mats = [loadmat(f.as_posix()) for f in files if '.mat' in f.suffixes]
    dicoms = [dcmread(f.as_posix()) for f in files if '.dcm' in f.suffixes]  # loaded dicom files
    embed()  # XXX you will drop into an interactive terminal in this scope


def mvp_main():
    bf, files = mvp()
    process_files(bf, files)

